unit DMath;

interface

const
  EPS = 10e-6;

function Min(A, B: Integer): Integer; overload;
function Max(A, B: Integer): Integer; overload;
function Min(A, B: Single): Single; overload;
function Max(A, B: Single): Single; overload;

type
  TVec2f = packed record
    class operator Add(A, B: TVec2f): TVec2f; inline;
    class operator Subtract(A, B: TVec2f): TVec2f; inline;
    class operator Multiply(A: TVec2f; S: single): TVec2f; overload; inline;
    class operator Multiply(S: single; A: TVec2f): TVec2f; overload; inline;
    class operator Multiply(A, B: TVec2f): TVec2f; overload; inline;
    class operator Divide(A, B: TVec2f): TVec2f; overload; inline;
    class operator Divide(A: TVec2f; S: Single): TVec2f; overload; inline;
    case integer of
      0: (X, Y: Single);
      1: (Arr: array[0..1] of Single);
  end;

  TVec2i = packed record
    class operator Implicit(A: TVec2i): TVec2f; inline;
    class operator Add(A, B: TVec2i): TVec2i; inline;
    class operator Subtract(A, B: TVec2i): TVec2i; inline;
    class operator Multiply(A: TVec2i; I: integer): TVec2i; overload; inline;
    class operator Multiply(A: TVec2i; S: single): TVec2f; overload; inline;
    class operator Multiply(A, B: TVec2i): TVec2i; overload; inline;
    class operator Divide(A, B: TVec2i): TVec2i; overload; inline;
    class operator Divide(A: TVec2i; I: Integer): TVec2i; overload; inline;
    case integer of
      0: (X, Y: integer);
      1: (Arr: array[0..1] of integer);
  end;

type
  TVec4f = packed record
    class operator Add(A, B: TVec4f): TVec4f; inline;
    class operator Subtract(A, B: TVec4f): TVec4f; inline;
    class operator Multiply(A: TVec4f; S: Single): TVec4f; overload; inline;
    class operator Multiply(S: Single; A: TVec4f): TVec4f; overload; inline;
    class operator Multiply(A, B: TVec4f): TVec4f; overload; inline;
    case integer of
      0: (X, Y, Z, W: Single);
      1: (Arr: array[0..3] of Single);
  end;

type
  TRectf = record
    A, B: TVec2f;
  end;
  TRect = TRectf;

type
  TLine2f = record
    N: TVec2f;
    D: Single;
  end;

function Vec2i(A, B: Integer): TVec2i;
function Len2i(A: TVec2i): Single;
function Dot2i(A, B: TVec2i): Integer;
function Round2f(V: TVec2f): TVec2i;

function Vec2f(A, B: Single): TVec2f; overload;
function Vec2f(V: TVec2i): TVec2f; overload;
function Len2f(A: TVec2f): Single;
function Dot2f(A, B: TVec2f): Single;
function Norm2f(V: TVec2f): TVec2f;

function Vec4f(A, B, C, D: Single): TVec4f;

function Rect(A, B: TVec2f): TRect;

function Line2f(Vec: TVec2f; S: Single): TLine2f; overload;
function Line2f(A, B: TVec2f): TLine2f; overload;
function Dist2f(Line: TLine2f; P: TVec2f): Single; overload;

implementation

{$Region 'MinMax'}

function Min(A, B: Integer): Integer;
begin
  if A < B then
    Min := A
  else
    Min := B;
end;

function Max(A, B: Integer): Integer;
begin
  if A > B then
    Max := A
  else
    Max := B;
end;

function Min(A, B: Single): Single;
begin
  if A < B then
    Min := A
  else
    Min := B;
end;

function Max(A, B: Single): Single;
begin
  if A > B then
    Max := A
  else
    Max := B;
end;

{$Endregion}

{$Region 'TVec2i'}

function Vec2i(A, B: Integer): TVec2i;
begin
  with Result do
  begin
    X := A;
    Y := B;
  end;
end;

class operator TVec2i.Add(A, B: TVec2i): TVec2i;
begin
  with Result do
  begin
    X := A.X + B.X;
    Y := A.Y + B.Y;
  end;
end;

class operator TVec2i.Subtract(A, B: TVec2i): TVec2i;
begin
  with Result do
  begin
    X := A.X - B.X;
    Y := A.Y - B.Y;
  end;
end;

class operator TVec2i.Multiply(A: TVec2i; I: integer): TVec2i;
begin
  with Result do
  begin
    X := I * A.X;
    Y := I * A.Y;
  end;
end;

class operator TVec2i.Multiply(A: TVec2i; S: single): TVec2f;
begin
  with Result do
  begin
    X := S * A.X;
    Y := S * A.Y;
  end;
end;

class operator TVec2i.Multiply(A, B: TVec2i): TVec2i;
begin
  with Result do
  begin
    X := A.X * B.X;
    Y := A.Y * B.Y;
  end;
end;

class operator TVec2i.Divide(A: TVec2i; B: TVec2i): TVec2i;
begin
  with Result do
  begin
    X := A.X div B.X;
    Y := A.Y div B.Y;
  end;
end;

class operator TVec2i.Divide(A: TVec2i; I: Integer): TVec2i;
begin
  with Result do
  begin
    X := A.X div I;
    Y := A.Y div I;
  end;
end;

function Round2f(V: TVec2f): TVec2i;
begin
  with Result do
  begin
    X := Round(V.X);
    Y := Round(V.Y);
  end;
end;

function Len2i(A: TVec2i): Single;
begin
  Result := Sqrt(Sqr(A.X) + Sqr(A.Y));
end;

function Dot2i(A, B: TVec2i): Integer;
begin
  Result := A.X * B.X + A.Y * B.Y;
end;

{$Endregion}

{$Region 'TVec2f'}

function Vec2f(A, B: Single): TVec2f;
begin
  with Result do
  begin
    X := A;
    Y := B;
  end;
end;

function Vec2f(V: TVec2i): TVec2f;
begin
  with Result do
    Vec2f(X, Y);
end;

class operator TVec2i.Implicit(A: TVec2i): TVec2f;
begin
  with Result do
  begin
    X := A.X;
    Y := A.Y;
  end;
end;

class operator TVec2f.Add(A, B: TVec2f): TVec2f;
begin
  with Result do
  begin
    X := A.X + B.X;
    Y := A.Y + B.Y;
  end;
end;

class operator TVEc2f.Subtract(A, B: TVec2f): TVec2f;
begin
  with Result do
  begin
    X := A.X - B.X;
    Y := A.Y - B.Y;
  end;
end;

class operator TVec2f.Multiply(A: TVec2f; S: single): TVec2f;
begin
  with Result do
  begin
    X := S * A.X;
    Y := S * A.Y;
  end;
end;

class operator TVec2f.Multiply(S: single; A: TVec2f): TVec2f;
begin
  with Result do
  begin
    X := S * A.X;
    Y := S * A.Y;
  end;
end;

class operator TVec2f.Multiply(A, B: TVec2f): TVec2f;
begin
  with Result do
  begin
    X := A.X * B.X;
    Y := A.Y * B.Y;
  end;
end;

class operator TVec2f.Divide(A: TVec2f; B: TVec2f): TVec2f;
begin
  with Result do
  begin
    X := A.X / B.X;
    Y := A.Y / B.Y;
  end;
end;

class operator TVec2f.Divide(A: TVec2f; S: Single): TVec2f;
begin
  with Result do
  begin
    X := A.X / S;
    Y := A.Y / S;
  end;
end;

function Len2f(A: TVec2f): Single;
begin
  Result := Sqrt(Sqr(A.X) + Sqr(A.Y));
end;

function Dot2f(A, B: TVec2f): Single;
begin
  Result := A.X * B.X  +  A.Y * B.Y;
end;

function Norm2f(V: TVec2f): TVec2f;
begin
  Result := V  / Len2f(V);
end;

{$Endregion}

{$Region 'TVec4f'}

function Vec4f(A, B, C, D: Single): TVec4f;
begin
  with Result do
  begin
    X := A;
    Y := B;
    Z := C;
    W := D;
  end;
end;

class operator TVec4f.Add(A, B: TVec4f): TVec4f;
begin
  with Result do
  begin
    X := A.X + B.X;
    Y := A.Y + B.Y;
    Z := A.Z + B.Z;
    W := A.W + B.W;
  end;
end;

class operator TVec4f.Subtract(A, B: TVec4f): TVec4f;
begin
  with Result do
  begin
    X := A.X - B.X;
    Y := A.Y - B.Y;
    Z := A.Z - B.Z;
    W := A.W - B.W;
  end;
end;

class operator TVec4f.Multiply(A: TVec4f; S: Single): TVec4f;
begin
  with Result do
  begin
    X := A.X * S;
    Y := A.Y * S;
    Z := A.Z * S;
    W := A.W * S;
  end;
end;

class operator TVec4f.Multiply(S: Single; A: TVec4f): TVec4f;
begin
  with Result do
  begin
    X := A.X * S;
    Y := A.Y * S;
    Z := A.Z * S;
    W := A.W * S;
  end;
end;

class operator TVec4f.Multiply(A, B: TVec4f): TVec4f;
begin
  with Result do
  begin
    X := A.X * B.X;
    Y := A.Y * B.Y;
    Z := A.Z * B.Z;
    W := A.W * B.W;
  end;
end;

{$Endregion}

{$Region 'TRect'}

function Rect(A, B: TVec2f): TRect;
begin
  Rect.A := A;
  Rect.B := B;
end;

{$Endregion}

{$Region 'TLine2f'}

function Line2f(Vec: TVec2f; S: Single): TLine2f;
begin
  with Result do
  begin
    N := Vec;
    D := S;
  end;
end;

function Line2f(A, B: TVec2f): TLine2f;
begin
  with Result do
  begin
    N := (B - A)/Len2f(B - A);
    N := Vec2f(- Result.N.Y, Result.N.X);
    D := - Dot2f(A, Result.N);
  end;
end;

function Dist2f(Line: TLine2f; P: TVec2f): Single; overload;
begin
  Result := Dot2f(Line.N, P) + Line.D;
end;

{$Endregion}

end.
