unit DGuiCommon;

interface

uses
  DUtils,
  DMath,
  DWinApi,
  DGuiCore;

type
TLabel = class(TWidgetDecorator)
private
  FColor: TVec4f;
  FFont: TGuiFont;
  FText: String;
  FAlign: TVec2f;
protected
  function GetFontName: String;
  procedure SetFont(NewFont: String); virtual;
  procedure SetText(NewText: String); virtual;
  procedure SetAlign(NewAlign: TVec2f); virtual;
public
  constructor Create(Widget: TWidget);
  destructor Destroy; override;
  procedure DrawLabel;
  function GetTextRect: TRect;
  property Color: TVec4f read FColor write FColor;
  property Font: String read GetFontName write SetFont;
  property Text: String read FText write SetText;
  property _Text: String read FText write SetText;
  property Align: TVec2f read FAlign write SetAlign;
end;

TGuiLabel = class(TWidget)
private
  FLabel: TLabel;
  procedure SetColor(NewColor: TVec4f); inline;
  function GetColor: TVec4f; inline;
  procedure SetFont(NewFont: String); inline;
  function GetFont: String; inline;
  procedure SetAlign(NewAlign: TVec2f); inline;
  function GetAlign: TVec2f; inline;
public
  constructor Create(Gui: TGui; Name, Style: String);
  destructor Destroy; override;
  procedure Render; override;
  property _Label: TLabel read FLabel;
  property Color: TVec4f read GetColor write SetColor;
  property Font: String read GetFont write SetFont;
  property Align: TVec2f read GetAlign write SetAlign;
end;

TGuiFonted = class(TWidget)
private
  FFont: TGuiFont;
  function GetFontName: String; inline;
  procedure SetFontName(NewFont: String); inline;
protected
  procedure OnFontChanged; virtual;
  property Font: TGuiFont read FFont write FFont;
public
  constructor Create(Gui: TGui; Name, Style: String);
  destructor Destroy; override;
  property FontName: String read GetFontName write SetFontName;
end;

TGuiCaption = class(TGuiFonted)
private
  FCaption: String;
  FAlign: TGuiAlign;
public
  procedure Render; override;
  property Caption: String read FCaption write FCaption;
  property Align: TGuiAlign read FAlign write FAlign;
end;

TGuiButton = class(TGuiCaption)
public
  FOnClick: TProcOfObj;
private
  procedure Render; override;
  function OnMouseDown(Ev: TWinApiMouse): Boolean; override;
  property OnClick: TProcOfObj read FOnClick write FOnClick;
end;

TGuiWindow = class(TGuiCaption)
  procedure Render; override;
end;

TGuiEdit = class(TGuiCaption)
private
  FOnEnter: TProcOfObj;
  FOnChange: TProcOfObj;
  FCursor: Integer;
  FDefault: String;
  procedure InsertCurrent(Text: String);
  procedure LeftCursor;
  procedure RightCursor;
  procedure Backspace;
public
  constructor Create(Gui: TGui; Name, Style: String);
  function OnMouseDown(Ev: TWinApiMouse): Boolean; override;
  function OnChar(Ev: TWinApiChar): Boolean; override;
  function OnKeyDown(Ev: TWinApiKey): Boolean; override;
  procedure Render; override;
  property _Default: String read FDefault write FDefault;
  property OnEnter: TProcOfObj read FOnEnter write FOnEnter;
  property OnChange: TProcOfObj read FOnChange write FOnChange;
end;

TGuiExtendedList = class(TGuiCaption)
private
  FCount: Integer;
  FCurrent: Integer;
  FSimpleHeight: Single;
  FExtendedHeight: Single;
  FColumnsCount: Integer;
protected
  procedure SetColumns(NewColumns: Integer);
  procedure DrawHeader(Pos: TVec2f); virtual;
  procedure DrawSimple(Pos: TVec2f; ID: Integer); virtual;
  procedure DrawExtended(Pos: TVec2f; ID: Integer); virtual;
public
  procedure Render; override;
  function OnMouseDown(Ev: TWinApiMouse): Boolean; override;
  function GetMouseActive: Integer;
  function GetHeight(ID: Integer): Single;
  property Count: Integer read FCount write FCount;
  property Current: Integer read FCurrent write FCurrent;
  property SimpleHeight: Single read FSimpleHeight write FSimpleHeight;
  property ExtendedHeight: Single read FExtendedHeight write FExtendedHeight;
  property ColumnsCount: Integer read FColumnsCount;
end;

implementation

constructor TLabel.Create(Widget: TWidget);
begin
  inherited Create(Widget);
  FFont := TGuiFont.Create(Widget.Style);
  FText := '';
  FAlign := Vec2f(0.5, 0.5);
  FColor := Widget.Style.PrimaryColor;
end;

destructor TLabel.Destroy;
begin
  FFont.Free;
end;

procedure TLabel.DrawLabel;
begin  
  //Style.DrawText(Widget.GlobalCenter, FAlign, FFont, FText);
  FFont.Draw(Widget.GlobalCenter, Vec2f(0.5, 0.5), FText);
end;

function TLabel.GetTextRect;
begin
  Result.A := Widget.GlobalCenter - FAlign * Vec2f(FFont.GetWidth(FText),
                                                   FFont.GetHeight(FText));
  Result.B := Result.A + Vec2f(FFont.GetWidth(FText),
                               FFont.GetHeight(FText));
end;

function TLabel.GetFontName: String;
begin
  Result := '<undefined>';
  if FFont <> nil then
    Result := FFont.Name;
end;

procedure TLabel.SetFont(NewFont: String);
begin
  FFont.Name := NewFont;
end;

procedure TLabel.SetText(NewText: String);
begin
  FText := NewText;
end;

procedure TLabel.SetAlign(NewAlign: TVec2f);
begin
  FAlign := NewAlign;
end;

procedure TGuiLabel.SetColor;
begin
  FLabel.Color := NewColor;
end;

function TGuiLabel.GetColor;
begin
  Result := FLabel.Color;
end;

procedure TGuiLabel.SetFont;
begin
  FLabel.Font := NewFont;
end;

function TGuiLabel.GetFont;
begin
  Result := FLabel.Font;
end;

procedure TGuiLabel.SetAlign;
begin
  FLabel.Align := NewAlign;
end;

function TGuiLabel.GetAlign;
begin
  Result := FLabel.Align;
end;

constructor TGuiLabel.Create(Gui: TGui; Name, Style: String);
begin
  inherited Create(Gui, Name, Style);
  FLabel := TLabel.Create(Self);
  FLabel.Color := Self.Style.PrimaryColor;
end;

destructor TGuiLabel.Destroy;
begin
  FLabel.Free;
  inherited Destroy;
end;

procedure TGuiLabel.Render;
begin
  Style.Color4f(Color);
  FLabel.DrawLabel;
end;

{$IFNDEF FLAG_FPC}{$REGION 'TGuiFonted'}{$ENDIF}
function TGuiFonted.GetFontName: String;
begin
  Result := FFont.Name;
end;

procedure TGuiFonted.SetFontName(NewFont: String);
begin
  if GetFontName <> NewFont then
    OnFontChanged;
  FFont.Name := NewFont;
end;

procedure TGuiFonted.OnFontChanged;
begin
end;

constructor TGuiFonted.Create(Gui: TGui; Name, Style: String);
begin
  inherited Create(Gui, Name, Style);
  FFont := TGuiFont.Create(Self.Style, 'default');
end;

destructor TGuiFonted.Destroy;
begin
  FFont.Free;
end;
{$IFNDEF FLAG_FPC}{$ENDREGION}{$ENDIF}

{$IFNDEF FLAG_FPC}{$REGION 'TGuiCaption'}{$ENDIF}
procedure TGuiCaption.Render;
begin
  Style.Color4f(Style.PrimaryColor);
  Font.Draw(GlobalRect, FAlign, FCaption);
end;
{$IFNDEF FLAG_FPC}{$ENDREGION}{$ENDIF}

{$IFNDEF FLAG_FPC}{$REGION 'TGuiButton'}{$ENDIF}
procedure TGuiButton.Render;
begin
  if IsMouseIn then
    Style.Color4f(Style.SecondaryActive)
  else
    Style.Color4f(Style.SecondaryNormal);
  Style.DrawRect(GlobalRect);
  if IsMouseIn then
    Style.Color4f(Style.PrimaryActive)
  else
    Style.Color4f(Style.PrimaryNormal);
  Style.SetLineWidth(2);
  Style.DrawRect(GlobalRect, False);  
  Font.Draw(GlobalRect, Align, Caption);
end;

function TGuiButton.OnMouseDown(Ev: TWinApiMouse): Boolean;
begin
  Result := False;
  if not IsGlobalPointIn(Ev.Pos) then
    Exit;
  if @FOnClick <> nil then
    FOnClick;
  Result := True;
end;
{$IFNDEF FLAG_FPC}{$ENDREGION}{$ENDIF}

{$IFNDEF FLAG_FPC}{$REGION 'TGuiWindow'}{$ENDIF}
procedure TGuiWindow.Render;
begin
  Style.Color4f(Style.SecondaryNormal);
  Style.DrawRect(GlobalRect);
  Style.Color4f(Style.PrimaryColor);
  Style.SetLineWidth(2);
  Style.DrawRect(GlobalRect, False);  
end;
{$IFNDEF FLAG_FPC}{$ENDREGION}{$ENDIF}

{$IFNDEF FLAG_FPC}{$REGION 'TGuiEdit'}{$ENDIF}
constructor TGuiEdit.Create(Gui: TGui; Name, Style: String);
begin
  inherited Create(Gui, Name, Style);
  Caption := '';
  FCursor := 1;
  FOnEnter := LostFocus;
end;

procedure TGuiEdit.InsertCurrent(Text: String);
  var
    S: String;
begin
  S := Caption;
  Insert(Text, S, FCursor);
  Caption := S;
  if @FOnChange <> nil then
    FOnChange;
  Inc(FCursor, Length(Text));
end;

procedure TGuiEdit.LeftCursor;
begin
  if FCursor > 1 then
    Dec(FCursor);
end;

procedure TGuiEdit.RightCursor;
begin
  if FCursor <= Length(Caption) then
    Inc(FCursor);
end;

procedure TGuiEdit.Backspace;
  var
    S: String;
begin
  if FCursor = 1 then
    Exit;
  S := Caption;
  if Length(S) = 0 then
    Exit;
  Delete(S, FCursor - 1, 1);
  Caption := S;
  LeftCursor;
end;

function TGuiEdit.OnMouseDown(Ev: TWinApiMouse): Boolean;
begin
  Result := False;
  if not IsGlobalPointIn(Ev.Pos) then
    Exit;
  SetFocus;
  Result := True;
end;

function TGuiEdit.OnChar(Ev: TWinApiChar): Boolean; 
begin
  Result := False;
  if not IsInFocus then
    Exit;
  if IsPrintable(Char(Ev.Key)) then
    InsertCurrent(Char(Ev.Key));
end;

function TGuiEdit.OnKeyDown(Ev: TWinApiKey): Boolean;
begin
  Result := False;
  if not IsInFocus then
    Exit;
  if Ev.Key = 8 then begin
    BackSpace;
  end else if (Ev.Key = KEY_ENTER) and Ev.First then begin
    if @FOnEnter <> nil then
      FOnEnter;
  end else if (Ev.Key in [Ord('v'), Ord('V')]) and Ev.Ctrl then begin
    InsertCurrent(Gui.Window.WinApi.GetFromClipboard);
  end else if (Ev.Key in [Ord('c'), Ord('C')]) and Ev.Ctrl then begin
    Gui.Window.WinApi.CopyToClipboard(Caption);
  end else if Ev.Key = KEY_LEFT then begin
    LeftCursor
  end else if Ev.Key = KEY_RIGHT then begin
    RightCursor
  end;
end;

procedure TGuiEdit.Render;
  var
    Text: String;
  procedure DrawCursor;
    var
      S: String;
      Pos: TVec2f;
  begin
    S := Copy(Caption + '=', 1, FCursor);
    with Style.PrimaryNormal do
      Style.Color4f(Vec4f(X, Y, Z, 0.2));
    Pos := Global + Vec2f(FFont.GetWidth(S) - FFont.GetSize.X + 
                           (Size.X - FFont.GetWidth(Caption))/2, 
                          Size.Y-FFont.GetSize.Y);
    Style.DrawRect(Rect(Pos + Vec2f(FFont.GetSize.X*1/4*0, 0), 
                        Pos + FFont.GetSize));
  end;
begin
  if IsMouseIn or IsInFocus then
    Style.Color4f(Style.MemoActive)
  else
    Style.Color4f(Style.MemoNormal);
  Style.DrawRect(GlobalRect);
  if IsMouseIn or IsInFocus then
    Style.Color4f(Style.PrimaryActive)
  else
    Style.Color4f(Style.PrimaryNormal);
  Style.SetLineWidth(2);
  Style.DrawRect(GlobalRect, False);  
  if (Caption = '') and not IsInFocus then
    Text := FDefault
  else
    Text := Caption;
  Style.Color4f(Style.PrimaryColor);
  Font.Draw(GlobalRect, FAlign, Text);
  if IsInFocus then
    DrawCursor;
end;
{$IFNDEF FLAG_FPC}{$ENDREGION}{$ENDIF}

{$IFNDEF FLAG_FPC}{$REGION 'TGuiExtendedList'}{$ENDIF}
procedure TGuiExtendedList.SetColumns(NewColumns: Integer);
begin
  FColumnsCount := NewColumns;
//  SetLength(FColumns, FColumnsCount);
end;

procedure TGuiExtendedList.DrawHeader(Pos: TVec2f);
begin
end;

procedure TGuiExtendedList.DrawSimple(Pos: TVec2f; ID: Integer);
begin
end;

procedure TGuiExtendedList.DrawExtended(Pos: TVec2f; ID: Integer);
begin
end;

procedure TGuiExtendedList.Render;
  var
    I: Integer;
    Y: Single;
begin
  Style.Color4f(Style.MemoNormal);
  Style.DrawRect(GlobalRect);
  Gui.Graph.EnableScissor(GlobalRect);
  DrawHeader(Vec2f(Global.X, Global.Y));
  Y := FSimpleHeight;
  for I := 0 to FCount - 1 do
    if I = FCurrent then begin
      DrawExtended(Vec2f(Global.X, Global.Y + Y), I);
      Y := Y + FExtendedHeight;
    end else begin
      DrawSimple(Vec2f(Global.X, Global.Y + Y), I);
      Y := Y + FSimpleHeight;
    end;
  Gui.Graph.DisableScissor;
  Style.Color4f(Style.PrimaryNormal);
  Style.SetLineWidth(2);
  Style.DrawRect(GlobalRect, False);  
end;

function TGuiExtendedList.OnMouseDown(Ev: TWinApiMouse): Boolean;
  var
    Active: Integer;
begin
  Result := False;
  if IsGlobalPointIn(Ev.Pos) then begin
    FCurrent := GetMouseActive;
  end;
end;

function TGuiExtendedList.GetMouseActive: Integer;
  var
    I: Integer;
    Y: Single;
begin
  Result := -1;
  if not IsMouseIn then
    Exit;
  Y := Global.Y + FSimpleHeight;
  for I := 0 to FCount - 1 do begin
    if Gui.Mouse.Y >= Y then begin
      if Gui.Mouse.Y <= Y + GetHeight(I) then begin
        Result := I;
        Exit;
      end;
    end else
      Exit;
    Y := Y + GetHeight(I);
  end;
end;

function TGuiExtendedList.GetHeight(ID: Integer): Single;
begin
  if ID = FCurrent then
    Result := FExtendedHeight
  else
    Result := FSimpleHeight;
end;
{$IFNDEF FLAG_FPC}{$ENDREGION}{$ENDIF}

end.
