unit DGuiConsole;

interface

uses
  DDebug,
  DUtils,
  DParser,
  DMath,
  DEvents,
  DWinApi,
  DGuiCore,
  DTML;

const
  ConstColor = '|#87.af.af.ff|';
  StandColor = '|#a8.a8.a8.ff|';
  ErrorColor = '|#87.00.00.ff|';

type
TGuiConsole = class;

TConsoleCommand = class
private
  FConsole: TGuiConsole;
  FName: String;
  FSummary: String;
  FColor: TVec4f;
protected
  procedure PrintTypeDescription; virtual;
  procedure PrintSummary; virtual;
public
  constructor Create(Name: String);
  procedure PrintDescription; virtual;
  procedure Process(Param: String); virtual;
  function GetColorName: String;
  property Console: TGuiConsole read FConsole;
  property Name: String read FName;
  property ColorName: String read GetColorName;
  property Summary: String read FSummary write FSummary;
  property Color: TVec4f read FColor write FColor;
end;
TListOfConsoleCommand = {$IFDEF FLAG_FPC}specialize{$ENDIF} TList<TConsoleCommand>;

TCommandHelp = class(TConsoleCommand)
public
  procedure PrintDescription; override;
  function PrintCommandList(Start: String): Boolean;
  procedure Process(Param: String); override;
end;

TCommandGroup = class(TConsoleCommand)
public
  procedure PrintDescription; override;
  procedure Process(Param: String); override;
end;

TConsoleProc = procedure(Param: String) of object;
TCommandProc = class(TConsoleCommand)
private
  FProc: TConsoleProc;
protected
  procedure PrintTypeDescription; override;
public
  constructor Create(Name: String; Proc: TConsoleProc = nil);
  procedure Process(Param: String); override;
  property Proc: TConsoleProc read FProc write FProc;
end;

TCommandValue = class(TConsoleCommand)
private
  FAdr: Pointer;
protected
  procedure SetValue(Value: String); virtual;
  function GetValue: String; virtual;
public
  constructor Create(Name: String; Adr: Pointer);
  procedure Process(Param: String); override;
  procedure PrintValue;
  property Adr: Pointer read FAdr write FAdr;
  //property Value: String read GetValue write SetValue;
end;

TCommandInteger = class(TCommandValue)
private
  FMin, FMax: Integer;
protected
  procedure PrintTypeDescription; override;
  function GetValue: String; override;
  procedure SetValue(Value: String); override;
public
  constructor Create(Name: String;
                     Adr: PInteger;
                     Min: Integer = Low(Integer);
                     Max: Integer = High(Integer));
end;

TCommandString = class(TCommandValue)
protected 
  procedure PrintTypeDescription; override;
  function GetValue: String; override;
  procedure SetValue(Value: String); override;
end;

TCommandFloat = class(TCommandValue)
private
  FMin, FMax: Single;
  FDigitals: Integer;
protected
  procedure PrintTypeDescription; override;
  function GetValue: String; override;
  procedure SetValue(Value: String); override;
public
  constructor Create(Name: String;
                     Adr: PSingle;
                     Min: Single = 0;
                     Max: Single = 1);
  property Digitals: Integer read FDigitals write FDigitals;
end;

TCommandBoolean = class(TCommandValue)
protected 
  procedure PrintTypeDescription; override;
  function GetValue: String; override;
  procedure SetValue(Value: String); override;
end;

TCommandSyn = class(TConsoleCommand)
private
  FOldName: String;
public
  constructor Create(OldName, NewName: String);
  procedure Process(Param: String); override;
  procedure PrintDescription; override;
  property OldName: String read FOldName;
end;

TConsoleLabel = record
  Pos: TVec2f;
  Text: String;
  Color: TVec4f;
  IsBackColor: Boolean;
  BackColor: TVec4f;
end;
TConsoleLine = array of TConsoleLabel;

TConsoleGroupsState = (cgs0, cgs1, cgsAll);
TConsoleGroupState = Boolean;
TConsoleGroup = record
  Name: String;
  Color: String;
  State: TConsoleGroupState;
end;

TColoredName = record
  Color: String;
  Name: String;
end;
TArrayOfColoredName = array Of TColoredName;

TGuiConsole = class(TWidget)
private
  FCur: record
    Text: String;
    Line: TConsoleLine;
    Cursor: Integer;
  end;
  FCommands: TListOfConsoleCommand;
  FFont: TGuiFont;
  FOpened: Boolean;
  FLines: array of TConsoleLine;
  FComp: record
    Enabled: Boolean;
    Active: Boolean;
    Comps: array Of String;
    Cur: Integer;
    Scroll: Integer;
    Width: Integer;
  end;
  FHistory: record
    Items: array of String;
    Cur: Integer;
  end;
  FGr: record
    State: TConsoleGroupsState;
    Groups: array of TConsoleGroup;
  end;
  FWidgets: TListOfWidget;
  procedure InsertCurrent(S: String);
  procedure BackSpace;
  procedure ClearCur;
  procedure ResetCursor;
  procedure SetCurrent(Text: String; NeedComp: Boolean = True);
  procedure LeftCursor;
  procedure RightCursor;
  procedure UpdateCurLine;
  procedure StartComp;
  procedure StopComp;
  procedure Complete;
  procedure RenderComp;
  procedure CompUp;
  procedure CompDown;
  procedure SetCurToComp;
  procedure UpdateCompScroll;
  procedure HistoryAdd(S: String);
  procedure HistoryUp;
  procedure HistoryDown;
  procedure UpdateLines;
  procedure AddLine(S: String);
  function GetFontName: String; inline;
  procedure SetFontName(NewFontName: String); inline;
  procedure SetOpened(NewOpened: Boolean);
  function FindCommand(Name: String): Integer; overload;
  procedure UpdateWidgetsPos;
public
  constructor Create(Gui: TGui; Style: String = 'default');
  destructor Destroy; override;
  procedure Update; override;
  procedure Render; override;
  procedure OnEvent(Ev: TEvent); override;
  function OnKeyDown(Ev: TWinApiKey): Boolean; override;
  function OnChar(Ev: TWinApiChar): Boolean; override;
  procedure AddCommand(Command: TConsoleCommand);
  procedure Process(S: String); overload;
  procedure Process(Name: String; Param: String); overload;
  procedure Print(S: ShortString); overload;
  procedure Print(S: Array Of Const); overload;
  function CalcWidth: Integer;
  function GetCommand(Name: String): TConsoleCommand; overload;
  procedure AddWidget(Widget: TWidget);
  procedure SetSummary(Name, Summary: String);
  function FindGroup(Name: String): Integer;
  function AddGroup(Name: String; Color: String = StandColor): Integer;
  procedure SetGroupsState(State: TConsoleGroupsState);
  procedure SetGroupState(Group: String; State: TConsoleGroupState);
  procedure SetGroupColor(Group: String; Color: String);
  procedure Print(Group, S: String); overload;
  procedure PrintList(List: TArrayOfColoredName);
  property Font: String read GetFontName write SetFontName;
  property Opened: Boolean read FOpened write SetOpened;
  property Commands: TListOfConsoleCommand read FCommands;
  property NamedCommands[Index: String]: TConsoleCommand read GetCommand; default;
end;

function ColorToHex(V: TVec4f): String;

implementation

{$IFDEF FLAG_FPC}
{uses
  lineinfo;}
{$ENDIF}

function GetLineInfo(P: Pointer): String;
  var
    Func, Source: ShortString;
    Line: LongInt;
begin
  Result := '';
  {$IFDEF FLAG_FPC}
  {lineinfo.GetLineInfo(LongWord(P), Func, Source, Line);
  Result := 'Line info: ';
  if Func <> '' then
    Result := Result + '|#87.00.d7|' + Func;
  if Source <> '' then begin
    if Func <> '' then
      Result := Result + StandColor + ', ';
    if Line <> 0 then
      Result := Result + StandColor + ' line ' + ConstColor + 
                             IntToStr(Line);
    Result := Result + StandColor + ' of ' + ConstColor + '"' + 
                           Source + '"';
  end;
  if Result = 'Line info: ' then
    Result := '';}
  {$ENDIF}
end;

function IsLiteral(C: Byte): Boolean;
begin
  Result := Char(C) in ['a'..'z', 'A'..'Z', 
                        ''..'', ''..'', 
                        '0'..'9', ' ', '#',
                        '.', ',', '?', '!',
                        '_', '-', '/', '\',
                        '[', ']', '{', '}',
                        '(', ')', '<', '>',
                        '$', '%', '&', '@',
                        ':', ';', '"', '''',
                        '+', '*', '=', '|',
                        '^'];
end;

function ColorToHex(V: TVec4f): String;
  var
    I: Integer;
begin
  Result := '';
  for I := 0 to 3 do
    Result := Result + IntToHex(Round(V.Arr[I]*255), 2) + '.';
  Result := StringDeletePost(Result, Length(Result));
end;

procedure TConsoleCommand.PrintTypeDescription;
begin
  if FSummary = '' then
    Console.Print(ColorName + StandColor + ' has not description');
end;

procedure TConsoleCommand.PrintSummary;
begin
  if FSummary <> '' then
    Console.Print(FSummary);
end;

constructor TConsoleCommand.Create(Name: String);
begin
  FName := Name;
  FConsole := nil;
  FSummary := '';
  FColor := Vec4f($00/255, $87/255, $D7/255, 1.0);
end;

procedure TConsoleCommand.PrintDescription;
begin
  //Console.Print(ColorName + FSummary);
  PrintTypeDescription;
  PrintSummary;
end;

procedure TConsoleCommand.Process(Param: String);
begin
end;

function TConsoleCommand.GetColorName: String;
begin
  Result := '|#' + ColorToHex(FColor) + '|' + Name;
end;

procedure TCommandHelp.PrintDescription; 
begin
  Console.Print('Coded by Doj in 2009');
  Console.Print('Type help             to get list of commands');
  Console.Print('Type help help        to get console overview');
  Console.Print('Type help examples    to get some common examples');
  Console.Print('Type help <command>   to get information about command');  
  Console.Print('Each command has name and param separeted by space: |#87.00.d7|<command> <param>');  
  Console.Print('Copy/Paste current line: Ctrl-C/Ctrl-V');  
  Console.Print('Console remembers last inserted commands, to use them press up/down arrows');  
  Console.Print('Console shows hints when you inserting command, press right arrow to complete');  
end;

function TCommandHelp.PrintCommandList;
  var
    Width: Integer;
    Columns: Integer;
    Column: String;
    Len: Integer;
    I, J: Integer;
    Commands: array of Integer;
begin
  Result := False;
  SetLength(Commands, 0);
  for I := 0 to Console.Commands.Last do
    if (Start = '') or (Pos(Start, Console.Commands[I].Name) <> 0) then begin
      SetLength(Commands, Length(Commands) + 1);
      Commands[High(Commands)] := I;
    end;
  if Length(Commands) = 0 then
    Exit;
  //if Start <> '' then
    Console.Print(['Found ', ConstColor, Length(Commands), StandColor, ' commands']);
  Len := 0;
  for I := 0 to High(Commands) do
    Len := Max(Len, Length(Console.Commands[Commands[I]].Name));
  Width := Console.CalcWidth;
  Columns := Trunc(Width/(Len + 1));
  Column := '';
  for I := 0 to High(Commands) do begin
    Column := Column + Console.Commands[Commands[I]].ColorName + 
              Space(Len - Length(Console.Commands[Commands[I]].Name) + 1);
    if (I + 1) mod Columns = 0 then begin
      Console.Print(Column);
      Column := '';
    end;
  end;
  if Column <> '' then
    Console.Print(Column);
  Result := True;
end;

procedure TCommandHelp.Process(Param: String);
  var
    Command: TConsoleCommand;
begin
  if Param = '' then
    PrintCommandList('')
  else begin
    Command := Console[Param];
    if Command = nil then begin
      if PrintCommandList(Param) then
        Exit;
      if Param = 'examples' then begin
        Console.Print('  :   ');
        Console.Print('    ,    ');
        Console.Print('  ,     ');
        Console.Print('   ,   ');
        Console.Print('    ,    help <command>');
        Console.Print('   :   |#87.00.d7|int 100    |#87.00.d7|float 0.5    |#87.00.d7|bool 1    |#87.00.d7|func');  
      end else
        Console.Print('|#87.00.00.ff|Command not found: "' + Param + '"')
    end else
      Command.PrintDescription;
  end;
end;

procedure TCommandGroup.PrintDescription; 
begin
  Console.Print('Type ' + ConstColor + 'gr                        ' + StandColor + 'to get groups state and groups list');
  Console.Print('Type ' + ConstColor + 'gr 0/1/all                ' + StandColor + 'to set state to disable/enable/all');
  Console.Print('Type ' + ConstColor + 'gr group 0/1/only         ' + StandColor + 'to set group state');
  Console.Print('Type ' + ConstColor + 'gr group #rr.gg.bb.aa     ' + StandColor + 'to set group color');
end;

procedure TCommandGroup.Process(Param: String);
  var
    Group, Action: String;
    I, J: Integer;
    S: String;
    List: TArrayOfColoredName;
begin
  if Param = '' then begin
    case Console.FGr.State of
      cgs0:   Console.Print('Groups ' + ConstColor + '0 (disabled)');
      cgs1:   Console.Print('Groups ' + ConstColor + '1 (normal)');
      cgsAll: Console.Print('Groups ' + ConstColor + 'all');
    end;
    SetLength(List, Length(Console.FGr.Groups));
    for I := 0 to High(List) do begin
      List[I].Color := Console.FGr.Groups[I].Color;
      List[I].Name := Console.FGr.Groups[I].Name;
    end;
    Console.PrintList(List);
  end else if Param = '0' then begin
    Console.SetGroupsState(cgs0);
    Console.Print('Groups ' + ConstColor + '0 (disabled)');
  end else if Param = '1' then begin
    Console.SetGroupsState(cgs1);
    Console.Print('Groups ' + ConstColor + '1 (normal)');
  end else if Param = 'all' then begin
    Console.SetGroupsState(cgsAll);
    Console.Print('Groups ' + ConstColor + 'all');
  end else begin
    ParseBinary(Param, ' ', Group, Action);
    Group := DeleteLeftSpaces(DeleteRightSpaces(Group));
    Action := DeleteLeftSpaces(DeleteRightSpaces(Action));
    if Pos(' ', Param) = 0 then
      Action := '';
    I := Console.FindGroup(Group);
    if Action = '' then begin
      if I = -1 then 
        Console.Print('Group "' + Group +'" does not exist')
      else begin
        Console.Print(Console.FGr.Groups[I].Color + 
                      Console.FGr.Groups[I].Name + StandColor +
                      ': ' + ConstColor +
                      IntToStr(Ord(Console.FGr.Groups[I].State)));
      end;
    end else begin
      if (Action <> '0')and(Action <> '1')and
         (Action <> 'only')and(Pos('#', Action) <> 1) then
        Console.Print(ErrorColor + 'Incorrect parameter: "' + Action + '"')
      else begin
        if I = -1 then
          I := Console.AddGroup(Group);
        if Action = '0' then
          Console.FGr.Groups[I].State := False
        else if Action = '1' then
          Console.FGr.Groups[I].State := True
        else if Action = 'only' then begin
          for J := 0 to High(Console.FGr.Groups) do
            Console.FGr.Groups[J].State := False;
          Console.FGr.Groups[I].State := True;
        end else begin
          Console.FGr.Groups[I].Color := '|' + Action + '|';
        end;
        Console.Print(Console.FGr.Groups[I].Color + 
                      Console.FGr.Groups[I].Name + StandColor +
                      ': ' + ConstColor +
                      IntToStr(Ord(Console.FGr.Groups[I].State)));       
      end;
    end;
    Exit;
  end;
end;

procedure TCommandProc.PrintTypeDescription;
  var
    Func, Source: ShortString;
    Line: Longint;
    LineInfo: String;
begin
  Console.Print('function ' + ColorName + StandColor + ': with parameter');
  LineInfo := GetLineInfo(@Proc);
  if LineInfo <> '' then
    Console.Print(LineInfo);
end;

constructor TCommandProc.Create(Name: String; Proc: TConsoleProc = nil);
begin
  inherited Create(Name);
  FProc := Proc;
end;

procedure TCommandProc.Process(Param: String);
begin
  if @FProc <> nil then
    FProc(Param)
  else
    Console.Print('Procedure "' + Name + '" is null and do nothing');
end;

procedure TCommandValue.SetValue(Value: String);
begin
  PrintValue;
end;

function TCommandValue.GetValue: String;
begin
  Result := '';
end;

constructor TCommandValue.Create(Name: String; Adr: Pointer);
begin
  inherited Create(Name);
  FAdr := Adr;
end;

procedure TCommandValue.Process(Param: String);
begin
  if Param <> '' then
    SetValue(Param)
  else
    PrintValue;
end;

procedure TCommandValue.PrintValue;
begin
  Console.Print(ColorName + '|#8c.8c.8c.ff|: |#87.af.af.ff|' + GetValue);
end;

procedure TCommandInteger.PrintTypeDescription;
begin
  Console.Print(['var ' + ColorName + '|#8c.8c.8c.ff|: |#87.af.87|integer|#8c.8c.8c.ff| in range [', ConstColor, FMin, '..', FMax, '|#8c.8c.8c.ff|]']);
end;

function TCommandInteger.GetValue: String;
begin
  Result := IntToStr(Integer(Adr^));
end;

procedure TCommandInteger.SetValue(Value: String);
  var
    Temp: Integer;
begin
  Temp := StrToInt(Value);
  if (Temp < FMin)or(Temp > FMax) then
    Console.Print(ToStr(['"', Name, '" must be in [', FMin, '..', FMax, ']']))
  else begin
    Integer(Adr^) := Temp;
    inherited;
  end;
end;

constructor TCommandInteger.Create(Name: String;
                                   Adr: PInteger;
                                   Min: Integer = Low(Integer);
                                   Max: Integer = High(Integer));
begin
  inherited Create(Name, Adr);
  FMin := Min;
  FMax := Max;
end;

procedure TCommandString.PrintTypeDescription;
begin
  Console.Print('var ' + ColorName + '|#8c.8c.8c.ff|: |#87.af.87|string');
end;

function TCommandString.GetValue: String;
begin
  Result := '"' + String(Adr^) + '"';
end;

procedure TCommandString.SetValue(Value: String);
begin
  String(Adr^) := Value;
  inherited;
end;

procedure TCommandFloat.PrintTypeDescription;
begin
  Console.Print(['var ' + ColorName + '|#8c.8c.8c.ff|: |#87.af.87|float|#8c.8c.8c.ff| in range [', ConstColor, FloatToStr(FMin, FDigitals), '..', FloatToStr(FMax, FDigitals), '|#8c.8c.8c.ff|] and prints with ', ConstColor, FDigitals, '|#8c.8c.8c.ff| digits']);
end;

function TCommandFloat.GetValue;
begin
  Result := FloatToStr(Single(Adr^), FDigitals);
end;

procedure TCommandFloat.SetValue;
  var
    Temp: Single;
begin
  Temp := StrToFloat(Value);
  if (Temp < FMin)or(Temp > FMax) then
    Console.Print(ToStr(['"', Name, '" must be in [', FMin, '..', FMax, ']']))
  else begin
    Single(Adr^) := Temp;
    inherited;
  end;
end;

constructor TCommandFloat.Create;
begin
  inherited Create(Name, Adr);
  FMin := Min;
  FMax := Max;
  FDigitals := 6;
end;

procedure TCommandBoolean.PrintTypeDescription;
begin
  Console.Print(['var ' + ColorName + '|#8c.8c.8c.ff|: |#87.af.87|boolean|#8c.8c.8c.ff| value must be 0 or 1']);
end;

function TCommandBoolean.GetValue;
begin
  Result := IntToStr(Ord(Boolean(Adr^)));
end;

procedure TCommandBoolean.SetValue;
begin
  Boolean(Adr^) := StrToInt(Value) <> 0;
  inherited;
end;

constructor TCommandSyn.Create(OldName, NewName: String);
  var
    Command: TConsoleCommand;
begin
  inherited Create(NewName);
  FOldName := OldName;
end;

procedure TCommandSyn.Process(Param: String);
  var
    Command: TConsoleCommand;
begin
  Command := Console[FoldName];
  if Command <> nil then
    Command.Process(Param);
end;

procedure TCommandSyn.PrintDescription; 
  var
    Command: TConsoleCommand;
begin
  Command := Console[FoldName];
  if Command <> nil then begin
    Color := Command.Color;  
    Console.Print(ColorName + StandColor + ' is synonim for ' + Command.ColorName);
    Command.PrintDescription;
  end else
    Console.Print(ColorName + StandColor + ' is synonim for ' + OldName);
end;

procedure TGuiConsole.InsertCurrent(S: String);
begin
  Insert(S, FCur.Text, FCur.Cursor);
  Inc(FCur.Cursor, Length(S));
  UpdateCurLine;
  if (Length(FCur.Text) > 1)and(Pos(' ', FCur.Text) = 0) then
    StartComp
  else
    StopComp;
end;

procedure TGuiConsole.BackSpace;
begin
  if FCur.Cursor = 1 then
    Exit;
  if FCur.Text = '' then
    Exit;
  Delete(FCur.Text, FCur.Cursor - 1, 1);
  if (Length(FCur.Text) > 1)and(Pos(' ', FCur.Text) = 0) then
    StartComp
  else
    StopComp;
  LeftCursor;
  UpdateCurLine;
end;

procedure TGuiConsole.ClearCur;
begin
  SetCurrent('');
end;

procedure TGuiConsole.ResetCursor;
begin
  FCur.Cursor := Length(FCur.Text) + 1;
end;

procedure TGuiConsole.SetCurrent;
begin
  FCur.Text := Text;
  ResetCursor;
  UpdateCurLine;
  if (Length(FCur.Text) > 1)and(Pos(' ', FCur.Text) = 0)and NeedComp then
    StartComp
  else
    StopComp;
end;

procedure TGuiConsole.LeftCursor;
begin
  if FCur.Cursor <= 1 then
    Exit;
  Dec(FCur.Cursor);
end;

procedure TGuiConsole.RightCursor;
begin
  if FCur.Cursor >= Length(FCur.Text) + 1 then
    Exit;
  Inc(FCur.Cursor);
end;

procedure TGuiConsole.UpdateCurLine;
begin
end;

procedure TGuiConsole.StartComp;
  var
    I: Integer;
begin
  if not FComp.Enabled then
    Exit;
  with FComp do begin
    Active := False;
    Cur := 0;
    Scroll := 0;
    SetLength(Comps, 0);
    Width := 8;
    for I := 0 to FCommands.Last do
      if Pos(FCur.Text, FCommands[I].Name) = 1 then begin
        SetLength(Comps, Length(Comps) + 1);
        Comps[High(Comps)] := FCommands[I].Name;
        Active := True;
        Width := Max(Width, Length(Comps[High(Comps)]));
      end;
  end;
end;

procedure TGuiConsole.StopComp;
begin
  FComp.Active := False;
end;

procedure TGuiConsole.Complete;
begin
  if (not FComp.Enabled) or (not FComp.Active) then
    Exit;
  SetCurrent(FComp.Comps[FComp.Cur] + ' ');
  FComp.Active := False;
end;

procedure TGuiConsole.RenderComp;
  var
    I: Integer;
    Count: Integer;
    P, S: TVec2f;
begin
  with FComp do begin
    if not Active then
      Exit;
    Count := Min(Length(Comps), 5);
    P := Local + Vec2f(0, Size.Y) - Vec2f(0, Count*FFont.GetSize.Y)
               + Vec2f(FFont.GetSize.X, -FFont.GetSize.Y);
    S := Vec2f(Width*FFont.GetSize.X, Count*FFont.GetSize.Y);
    Style.Color4f(Vec4f($2E/255, $34/255, $36/255, 1));
    Style.DrawRect(Rect(P, P + S));
    Style.Color4f(Vec4f($EE/255, $EE/255, $EC/255, 1));
    for I := 0 to Count - 1 do begin
      if Scroll + I = Cur then begin
        Style.Color4f(Vec4f($8C/255, $8C/255, $8C/255, 1));
        Style.DrawRect(Rect(P + Vec2f(0, I*FFont.GetSize.Y), 
                            P + Vec2f(0, I*FFont.GetSize.Y) +
                                Vec2f(Width*FFont.GetSize.X, FFont.GetSize.Y)));
        Style.Color4f(Vec4f($8C/255, $24/255, $26/255, 1))
      end else
        Style.Color4f(Vec4f($EE/255, $EE/255, $EC/255, 1));
      FFont.Draw(P + Vec2f(0, I*FFont.GetSize.Y), Comps[Scroll + I]);
    end;
  end;
end;

procedure TGuiConsole.SetCurToComp;
begin
  if (not FComp.Enabled) or (not FComp.Active) then
    Exit;
  FCur.Text := FComp.Comps[FComp.Cur] + ' ';
  ResetCursor;
  //SetCurrent(FComp.Comps[FComp.Cur] + ' ');
  //StartComp;
end;

procedure TGuiConsole.UpdateCompScroll;
begin
  with FComp do begin
    if Cur > Scroll + 4 then
      Scroll := Max(0, Cur - 4);
    if Cur < Scroll then
      Scroll := Cur;
  end;
end;

procedure TGuiConsole.HistoryAdd(S: String);
  var
    I, J: Integer;
begin
  with FHistory do begin
    for I := 0 to High(Items) do
      if Items[I] = S then begin
        for J := I to High(Items) - 1 do
          Items[J] := Items[J + 1]; 
        SetLength(Items, Length(Items) - 1);
        Break;
      end;
    if Length(Items) >= 30 then begin
      for I := 0 to High(Items) - 1 do
        Items[I] := Items[I + 1];
    end else 
      SetLength(Items, Length(Items) + 1);
    Items[High(Items)] := S;
    Cur := High(Items);
  end;
end;

procedure TGuiConsole.HistoryUp;
begin
  if Length(FHistory.Items) = 0 then
    Exit;
  if FCur.Text = FHistory.Items[FHistory.Cur] then
    FHistory.Cur := Max(0, FHistory.Cur - 1);
  SetCurrent(FHistory.Items[FHistory.Cur], False);
end;

procedure TGuiConsole.HistoryDown;
begin
  if Length(FHistory.Items) = 0 then
    Exit;
  if FCur.Text = FHistory.Items[FHistory.Cur] then
    FHistory.Cur := Min(High(FHistory.Items), FHistory.Cur + 1);
  SetCurrent(FHistory.Items[FHistory.Cur], False);
end;

procedure TGuiConsole.CompUp;
begin
  with FComp do begin
    if (not Enabled) or (not Active) then
      Exit;
    Cur := Cur - 1;
    if Cur < 0 then
      Cur := High(Comps);
    UpdateCompScroll;
  end;
  SetCurToComp;
end;

procedure TGuiConsole.CompDown;
begin
  with FComp do begin
    if (not Enabled) or (not Active) then
      Exit;
    Cur := (Cur + 1) mod Length(Comps);
    UpdateCompScroll;
  end;
  SetCurToComp;
end;

procedure TGuiConsole.UpdateLines;
begin
  SetLength(FLines, Trunc(Size.Y/FFont.GetSize.Y - 1));
end;

procedure TGuiConsole.AddLine(S: String);
  var
    I: Integer;
    X: Single;
    DTML: TDTML;
    Cur: TConsoleLabel;
    function StrToColor(S: String): TVec4f;
      var
        Components: TArrayOfString;
        I: Integer;
    begin
      ParseList(S, '.', Components);
      Result := Vec4f(0, 0, 0, 1);
      for I := 0 to Min(3, High(Components)) do
        Result.Arr[I] := HexToInt(Components[I])/255;
    end;
    procedure DoCommand(S: String);
    begin
      if Pos('#', S) = 1 then begin
        Cur.Color := StrToColor(Copy(S, 2, Length(S) - 1));
      end else if Pos('b#', S) = 1 then begin
        Cur.IsBackColor := Length(S) > 2;
        if Cur.IsBackColor then
          Cur.BackColor := StrToColor(Copy(S, 3, Length(S) - 2));
      end;
    end;
begin
  for I := 0 to High(FLines) - 1 do
    FLines[I] := FLines[I + 1];
  ParseDTML(S, DTML);
  SetLength(FLines[High(FLines)], 0);
  Cur.Color := Vec4f($A8/255, $A8/255, $A8/255, 1.0);
  Cur.IsBackColor := False;
  Cur.Pos.X := 0;
  for I := 0 to High(DTML) do
    if DTML[I].Command then begin
      DoCommand(DTML[I].Text);
    end else begin
      SetLength(FLines[High(FLines)], Length(FLines[High(FLines)]) + 1);
      with FLines[High(FLines)][High(FLines[High(FLines)])] do begin
        Text := DTML[I].Text;
        Color := Cur.Color;
        IsBackColor := Cur.IsBackColor;
        BackColor := Cur.BackColor;
        Pos.X := Cur.Pos.X;
        Pos.Y := 0;
        Cur.Pos.X := Cur.Pos.X + FFont.GetWidth(Text);
      end;
    end;
end;

function TGuiConsole.GetFontName: String;
begin
  Result := FFont.Name;
end;

procedure TGuiConsole.SetFontName(NewFontName: String);
begin
  FFont.Name := NewFontName;
end;

procedure TGuiConsole.SetOpened(NewOpened: Boolean);
begin
  FOpened := NewOpened;
  if FOpened then begin
    Local := Vec2f(0, 0);
  end else
    Local := Vec2f(0, -Size.Y);
end;

function TGuiConsole.FindCommand(Name: String): Integer;
begin
  Result := FCommands.Last;
  while Result >= 0 do
    if FCommands[Result].Name = Name then
      Exit
    else
      Dec(Result);
end;

procedure TGuiConsole.UpdateWidgetsPos;
  var
    I: Integer;
    Y: Single;
begin
  Y := Size.Y - FFont.GetSize.Y;
  for I := FWidgets.Last downto 0 do begin
    Y := Y - FWidgets[I].Size.Y;
    if FWidgets[I].Visible then
      FWidgets[I].Local := Vec2f(Size.X - FWidgets[I].Size.X, Y);
  end;
end;

constructor TGuiConsole.Create;
begin
  inherited Create(Gui, 'console', Style);
  AlwaysOnTop := True;
  FFont := TGuiFont.Create(Self.Style, 'default');
  FCommands := TListOfConsoleCommand.Create;
  //FLines := TListOfString.Create;
  SetLength(FLines, 0);
  Parent := Gui.Root;
  Size := Vec2f(Parent.Size.X, Parent.Size.Y/3);
  Local := Vec2f(0, -Size.Y);
  FOpened := False;
  ClearCur;
  UpdateLines;
  AddCommand(TCommandHelp.Create('help'));
  AddCommand(TCommandGroup.Create('gr'));
  FComp.Enabled := True;
  StopComp;
  SetLength(FHistory.Items, 0);
  FHistory.Cur := 0;
  FWidgets := TListOfWidget.Create;

  FGr.State := cgs1;
  AddGroup('error', ErrorColor);

  Print('Type help help        to get console overview');
end;

destructor TGuiConsole.Destroy;
  var
    I: Integer;
begin
  //FLines.Free;
  FWidgets.Free;
  for I := 0 to FCommands.Last do
    FCommands[I].Free;
  FCommands.Free;
  FFont.Free;
end;

procedure TGuiConsole.Update;
begin
  if FOpened then begin
    Local := Vec2f(0, Min(0, Local.Y + 15));
  end else
    Local := Vec2f(0, Max(-Size.Y, Local.Y - 15));
end;

procedure TGuiConsole.Render;
  var
    I, J: Integer;
  procedure DrawCursor;
    var
      S: String;
  begin
    S := Copy('>' + FCur.Text, 1, FCur.Cursor);
    Style.Color4f(Vec4f($A8/255, $A8/255, $A8/255, 0.2)); // Standart
    Style.DrawRect(Rect(Local + Vec2f(FFont.GetWidth(S) 
                                      + FFont.GetSize.X*1/4, 
                                      Size.Y-FFont.GetSize.Y), 
                        Local + Vec2f(FFont.GetWidth(S), 
                                      Size.Y-FFont.GetSize.Y) + FFont.GetSize));
  end;
begin
  Style.Color4f(Vec4f(0, 0, 0, 0.5));
  Style.DrawRect(GlobalRect);
  
  if FComp.Active then begin
    Style.Color4f(Vec4f($87/255, $87/255, $AF/255, 1.0));
    FFont.Draw(Local + Vec2f(0, Size.Y), Vec2f(0, 1), '>' + FComp.Comps[FComp.Cur]);
  end;

  Style.Color4f(Vec4f($87/255, $00/255, $D7/255, 1.0)); // preproc
  Style.Color4f(Vec4f($87/255, $87/255, $AF/255, 1.0)); // comment
  Style.Color4f(Vec4f($87/255, $AF/255, $AF/255, 1.0)); // constant
  Style.Color4f(Vec4f($AF/255, $87/255, $00/255, 1.0)); // TODO

  Style.Color4f(Vec4f($00/255, $AF/255, $AF/255, 1.0)); // Identifier
  Style.Color4f(Vec4f($00/255, $AF/255, $87/255, 1.0)); // Special
  Style.Color4f(Vec4f($00/255, $87/255, $D7/255, 1.0)); // Statement

  Style.Color4f(Vec4f($87/255, $AF/255, $87/255, 1.0)); // Type
  Style.Color4f(Vec4f($A8/255, $A8/255, $A8/255, 1.0)); // Standart

  FFont.Draw(Local + Vec2f(0, Size.Y), Vec2f(0, 1), '>' + FCur.Text);
  DrawCursor;
  Style.Color4f(Vec4f($A8/255, $A8/255, $A8/255, 1.0)); // Standart
  for I := 0 to High(FLines) do
    for J := 0 to High(FLines[I]) do 
      with FLines[I][J] do begin
        Style.Color4f(Color);
        FFont.Draw(Local+Vec2f(Pos.X, Size.Y), Vec2f(0, Length(FLines)-I+1), Text);
      end;
  RenderComp;
end;

procedure TGuiConsole.OnEvent(Ev: TEvent);
begin
  {if Ev is TWinApiClipboard then
    with TWinApiClipboard(Ev) do
      if Action = caPaste then
        InsertCurrent(Gui.Window.WinApi.GetFromClipboard);}
end;

function TGuiConsole.OnKeyDown(Ev: TWinApiKey): Boolean; 
begin
  //if Ev.Ctrl then
  //  Print('|#ff.00.00|Ctrl');
  Result := False;
  if (Ev.Key = KEY_TILDE) and Ev.First then
    FOpened := not FOpened;
  if not FOpened then
    Exit;
  if Ev.Key = 8 then begin
    BackSpace;
  end else if (Ev.Key = KEY_ENTER) and Ev.First then begin
    Process(FCur.Text);
    HistoryAdd(FCur.Text);
    ClearCur;
  end else if (Ev.Key in [Ord('v'), Ord('V')]) and Ev.Ctrl then begin
    InsertCurrent(Gui.Window.WinApi.GetFromClipboard);
  end else if (Ev.Key in [Ord('c'), Ord('C')]) and Ev.Ctrl then begin
    Gui.Window.WinApi.CopyToClipboard(FCur.Text);
  end else if Ev.Key = KEY_LEFT then begin
    LeftCursor
  end else if FComp.Active then begin
    if Ev.Key = KEY_DOWN then begin
      CompDown
    end else if Ev.Key = KEY_UP then begin
      CompUp;
    end else if Ev.Key = KEY_RIGHT then begin
      Complete;
    end;
  end else begin
    if Ev.Key = KEY_RIGHT then
      RightCursor;
    if Ev.Key = KEY_DOWN then begin
      HistoryDown
    end else if Ev.Key = KEY_UP then begin
      HistoryUp;
    end;
  end;
end;

function TGuiConsole.OnChar(Ev: TWinApiChar): Boolean;
begin
  Result := False;
  if not FOpened then
    Exit;
  if IsLiteral(Ev.Key) then
    InsertCurrent(Chr(Ev.Key));
end;

procedure TGuiConsole.AddCommand(Command: TConsoleCommand);
  var
    I: Integer;
begin
  if Command.Console <> nil then
    Exit;
  Command.FConsole := Self;
  I := FindCommand(Command.Name);
  if I <> -1 then begin
    FCommands[I].Free;
    FCommands[I] := Command;
  end else
    FCommands.Add(Command);
end;

procedure TGuiConsole.Process(S: String); 
  var
    Command, Param: String;
begin
  if Pos(' ', S) <> 0 then 
    ParseBinary(S, ' ', Command, Param)
  else begin
    Command := S;
    Param := '';
  end;
  Command := DeleteLeftSpaces(DeleteRightSpaces(Command));
  Param := DeleteLeftSpaces(DeleteRightSpaces(Param));
  Process(Command, Param);
end;

procedure TGuiConsole.Process(Name: String; Param: String);
  var
    I: Integer;
begin
  I := FindCommand(Name);
  if I = -1 then
    //$87/255, $00/255, $D7/255, 1.0)
    Print('|#87.00.00.ff|Command not found: "' + Name + '"')
  else
    FCommands[I].Process(Param);
end;

procedure TGuiConsole.Print(S: ShortString);
begin
  //SDebug.Log(S);
  AddLine(S);
end;

procedure TGuiConsole.Print(S: Array Of Const);
begin
  Print(ToStr(S));
end;

function TGuiConsole.CalcWidth;
begin
  Result := Trunc(Size.X / FFont.GetSize.X);
end;

function TGuiConsole.GetCommand(Name: String): TConsoleCommand;
  var
    I: Integer;
begin
  Result := nil;
  for I := 0 to FCommands.Last do
    if FCommands[I].Name = Name then begin
      Result := FCommands[I];
      Exit;
    end;
end;

procedure TGuiConsole.AddWidget(Widget: TWidget);
begin
  FWidgets.Add(Widget);
  Widget.Parent := Self;
  UpdateWidgetsPos;
end;

procedure TGuiConsole.SetSummary(Name, Summary: String);
  var
    Command: TConsoleCommand;
begin
  Command := GetCommand(Name);
  if Command <> nil then
    Command.Summary := Summary;
end;

function TGuiConsole.FindGroup(Name: String): Integer;
begin
  Result := High(FGr.Groups);
  while Result >= 0 do
    if FGr.Groups[Result].Name = Name then
      Exit
    else
      Dec(Result);
end;

function TGuiConsole.AddGroup(Name: String; Color: String = StandColor): Integer;
begin
  with FGr do begin
    Result := FindGroup(Name);
    if Result = -1 then begin
      SetLength(Groups, Length(Groups) + 1);
      Result := High(Groups);
    end;
    Groups[Result].Name := Name;
    Groups[Result].Color := Color;
    Groups[Result].State := False;
  end;
end;

procedure TGuiConsole.SetGroupsState(State: TConsoleGroupsState);
begin
  FGr.State := State;
end;

procedure TGuiConsole.SetGroupState(Group: String; State: TConsoleGroupState);
  var
    I: Integer;
begin
  I := FindGroup(Group);
  if I = -1 then
    I := AddGroup(Group);
  FGr.Groups[I].State := State;
end;

procedure TGuiConsole.SetGroupColor(Group: String; Color: String);
  var
    I: Integer;
begin
  I := FindGroup(Group);
  if I = -1 then
    I := AddGroup(Group);
  FGr.Groups[I].Color := Color;
end;

procedure TGuiConsole.Print(Group, S: String);
  var
    I: Integer;
begin
  if FGr.State = cgs0 then
    Exit;
  I := FindGroup(Group);
  if I = -1 then
    I := AddGroup(Group);
  if FGr.Groups[I].State or (FGr.State = cgsAll) then
    Print(FGr.Groups[I].Color + '[' + FGr.Groups[I].Name + '] ' + S);
end;

procedure TGuiConsole.PrintList(List: TArrayOfColoredName);
  var
    Width: Integer;
    Columns: Integer;
    Column: String;
    Len: Integer;
    I, J: Integer;
begin
  Len := 0;
  for I := 0 to High(List) do
    Len := Max(Len, Length(List[I].Name));
  Width := CalcWidth;
  Columns := Trunc(Width/(Len + 1));
  Column := '';
  for I := 0 to High(List) do begin
    Column := Column + List[I].Color + List[I].Name + 
              Space(Len - Length(List[I].Name) + 1);
    if (I + 1) mod Columns = 0 then begin
      Print(Column);
      Column := '';
    end;
  end;
  if Column <> '' then
    Print(Column);
end;

end.
