unit DMap;

interface

uses
  DCore,
  DUtils,
  DDebug,
  DGraph,
  DMath,
  DIntersection,
  
  DLevelData,
  DXML,
  DXMLExecutor;

type
TMap = class;

TSnowParticle = record
  Pos: TVec2f;
  Start: Integer;
end;

TBlock = record
  Tex: TTexture;
  Rect: TRect;
  Tile: TVec2f;
  Blend: TBlendMode;
  Color: TVec4f;
  Pos1, Pos2: TVec2f;
  Speed: Single; // -    
end;

TEntry = class(TStreamable)//{{{
private
  FMap: TMap;
  FName: String;
  FRect: TRect;
  FActive: Boolean;
protected
  procedure Stream; override;
public
  constructor Create(Map: TMap; Name: string; Rect: TRect);
  procedure DoActivate; virtual;
  procedure DoDeactivate; virtual;
  procedure Activate;
  procedure Deactivate;
  procedure Render; virtual;
  procedure Update; virtual;
  property Map: TMap read FMap;
  property Name: string read FName;
  property Rect: TRect read FRect;
  property Active: Boolean read FActive;
end;
//}}}
TEntryTrap = class(TEntry)//{{{
public
  procedure Update; override;
end;
//}}}
TEntryTexture = class(TEntry)//{{{
private
  FTexture: TTexture;
  FBlend: TBlendMode;
  FColor: TVec4f;
protected
  procedure Stream; override;
public
  constructor Create(Map: TMap; Name: string; Rect: TRect; Blend: TBlendMode;
    Color: TVec4f);
  destructor Destroy; override;
  procedure Render; override;
end;
//}}}
TEntryText = class(TEntry)//{{{
private
  FText: String;
  FPos: TVec2f;
  FAlign: TVec2f;
  FColor: TVec4f;
protected
  procedure Stream; override;
public
  constructor Create(Pos, Align: TVec2f; Color: TVec4f; Text: String);
  procedure Render; override;
end;
//}}}
TEntryJump2Pattern = class(TEntryTrap)//{{{
private
  FPattern: Integer;
  FJumped: Boolean;
protected
  procedure Stream; override;
public
  constructor Create(Map: TMap; Name: string; Rect: TRect; Pattern: Integer);
  procedure DoActivate; override;
end;
//}}}
TEntryMusic = class(TEntryTrap)//{{{
private
  FPlayed: Boolean;
protected
  procedure Stream; override;
public
  constructor Create(Name: String; Rect: TRect);
  procedure DoActivate; override;
end;
//}}}
TEntryTag = class(TEntryTrap)//{{{
private
  FTag: string;
protected
  procedure Stream; override;
public
  constructor Create(Map: TMap; Name: string; Rect: TRect; Tag: String);
  procedure Update; override;  
end;
//}}}
TEntrySave = class(TEntryTrap)//{{{
private
  FTexture: TTexture;
  FSaved: Boolean;
protected
  procedure Stream; override;
public
  constructor Create(Map: TMap; X, Y: Single; Name: String);
  procedure DoActivate; override;
  procedure Render; override;
end;
//}}}
TEntryBossDoj = class(TEntry)//{{{
private
  FCreated: Boolean;
protected
  procedure Stream; override;
public
  constructor Create(Map: TMap; Name: string; Rect: TRect);
  procedure DoActivate; override;
end;
//}}}
TEntryRandomRespawn = class(TEntry)//{{{
private
  FEPS: Single;
protected
  procedure Stream; override;
public
  constructor Create(Rect: TRect; EPS: Single);
  procedure Put(Pos: TVec2f); virtual;
  procedure Update; override;
end;
//}}}
TEntryWheelRespawn = class(TEntry)//{{{
private
  FLevel: Integer;
  FVel: TVec2f;
protected
  procedure Stream; override;
public
  constructor Create(Name: string; Rect: TRect; Level: Integer; Vel: TVec2f);
  procedure DoActivate; override;
end;//}}}
TSimpleRespawn = class(TEntryRandomRespawn)//{{{
  procedure Put(Pos: TVec2f); override;
end;//}}}
TEntryRespawnRect = class(TEntry)//{{{
private
  FWidth, FHeight: Integer;
  FVel: TVec2f;
protected
  procedure Stream; override;
public
  constructor Create(Name: String; Pos, Vel: TVec2f; W, H: Integer);
  procedure DoActivate; override;
end;
//}}}
TEntryTimer = class(TEntry)//{{{
private
  FTime: Integer;
  FInterval: Integer;
  FTag: String;
protected
  procedure Stream; override;
public
  constructor Create(Interval: Integer; Tag: String);
  procedure Update; override;
end;
//}}}
TEntryTotalWin = class(TEntryTrap)//{{{
  procedure DoActivate; override;
end;//}}}
// TMap//{{{
TMapCollision = (mcNo, mcUp, mcSide);

TMap = class(TStreamable)
private
  FHeap: TTexture;
  FBlocks: array of TBlock;
  FEntries: array of TEntry;
  FHeapColor: TVec4f;
  FSnow: array of TSnowParticle;
  
  FXMLPos: TVec2f;
  FXMLSize: TVec2f;
  FXMLTex: TString;
  FXMLHeap: TString;
  FXMLBlend: TBlendMode;
  FXMLColor: TVec4f;
  FXMLTile: TVec2f;

  function GetBlock(Index: Integer): TBlock;
  function GetEntry(Index: Integer): TEntry;
  procedure SortBlocks;
protected
  procedure Stream; override;
public
  constructor Create(Data: TData);
  //procedure LoadBlocks(Blocks: array of TBlockData);
  procedure LoadEntries;
  procedure Render;
  procedure Update;
  procedure UpdateEntries(Player: TRect);
  function Collide(Rect: TRect; 
                   Block: Integer;
                   var Delta: TVec2f): TMapCollision;
  function Higher(Rect: TRect; Block: Integer): Boolean;
  function GetLevel(Rect: TRect): Integer;
  procedure AddBlock(Block: TBlock);
  procedure AddEntry(Entry: TEntry);
  function BlocksCount: Integer; inline;
  function EntriesCount: Integer; inline;
  
  procedure XMLBlocks(Exe: TXMLExecutor; Item: PXMLItem; Into: Boolean);
  procedure XMLBlock(Exe: TXMLExecutor; Item: PXMLItem; Into: Boolean);

  property Blocks[Index: Integer]: TBlock read GetBlock;
  property Entries[Index: Integer]: TEntry read GetEntry;
end;
//}}}

implementation

uses
  DGame,
  DBossDoj;

//{{{ TEntry
procedure TEntry.Stream;
begin
  inherited;
  FMap := Game.Map;
  Stream(FName);
  Stream(@FRect, SizeOf(FRect));
  Stream(@FActive, SizeOf(FActive));
end;

constructor TEntry.Create;
begin
  FMap := Map;
  FName := Name;
  FRect := Rect;
  FActive := False;
end;

procedure TEntry.DoActivate;
begin
end;

procedure TEntry.DoDeactivate; 
begin
end;

procedure TEntry.Activate;
begin
  if not FActive then
    DoActivate;
  FActive := True;
end;

procedure TEntry.Deactivate;
begin
  if FActive then
    DoDeactivate;
  FActive := False;
end;

procedure TEntry.Render;
begin
end;

procedure TEntry.Update;
begin
end;
//}}}
//{{{ TEntryTrap
procedure TEntryTrap.Update; 
begin
  if Intersect(Rect, Game.Player.Rect) then
    Activate
  else
    Deactivate;
end;

//}}}
//{{{ TEntryTexture
procedure TEntryTexture.Stream;
begin
  inherited;
  if Loading then
    FTexture := Core.Graph.LoadTex(Data.ReadStr)
  else
    Data.WriteStr(Core.Graph.GetTexFileName(FTexture));
  Stream(@FBlend, SizeOf(FBlend));
  Stream(@FColor, SizeOf(FColor));
end;

constructor TEntryTexture.Create;
begin
  inherited Create(Map, Name, Rect);
  FTexture := Core.Graph.LoadTex(Name);
  FBlend := Blend;
  FColor := Color;
end;

destructor TEntryTexture.Destroy;
begin
end;

procedure TEntryTexture.Render; 
begin
  with Core.Graph do begin
    SetBlendMode(FBlend);
    Color4f(FColor);
    SetTexture(FTexture);
    DrawTexture(Rect);
    SetTexture(-1);
  end;
end;
//}}}
//{{{ TEntryText
procedure TEntryText.Stream;//{{{
begin
  inherited;
  Stream(@FPos, SizeOf(FPos));
  Stream(@FAlign, SizeOf(FAlign));
  Stream(@FColor, SizeOf(FColor));
  Stream(FText);
end;
//}}}
constructor TEntryText.Create(Pos, Align: TVec2f; Color: TVec4f; Text: String);//{{{
begin
  inherited Create(Game.Map, Text, DMath.Rect(Pos-Vec2f(10, 10), 
                                              Pos + Vec2f(10, 10)));
  FPos := Pos;
  FAlign := Align;
  FColor := Color;
  FText := Text;
end;
//}}}
procedure TEntryText.Render;//{{{
begin
  Core.Graph.Color4f(FColor);
  Game.DrawText(FPos.X, FPos.Y, FAlign.X, FAlign.Y, FText);
end;
//}}}
//}}}
//{{{ TEntryJump2Pattern
procedure TEntryJump2Pattern.Stream;
begin
  inherited;
  Stream(@FPattern, SizeOf(FPattern));
end;

constructor TEntryJump2Pattern.Create;
begin
  inherited Create(Map, Name, Rect);
  FPattern := Pattern;
  FJumped := False;
end;

procedure TEntryJump2Pattern.DoActivate;
begin
  if FJumped then
    Exit;
  FJumped := True;
  Core.Sound.MusicPattern := FPattern;
end;
//}}}
//{{{ TEntryMusic
procedure TEntryMusic.Stream;
begin
  inherited;
  Stream(@FPlayed, SizeOf(FPlayed));
end;

constructor TEntryMusic.Create;
begin
  inherited Create(Game.Map, Name, Rect);
  FPlayed := False;
end;

procedure TEntryMusic.DoActivate;
begin
  if FPlayed then
    Exit;
  FPlayed := True;
  Core.Sound.PlayMusic(Name);
end;
//}}}
//{{{ TEntryTag
procedure TEntryTag.Stream;
begin
  inherited;
  Stream(FTag);
end;

constructor TEntryTag.Create;
begin
  inherited Create(Map, Name, Rect);
  FTag := Tag;
end;

procedure TEntryTag.Update;
  var
    I: Integer;
begin
  inherited;
  for I := 0 to Map.EntriesCount - 1 do
    if Map.Entries[I].Name = FTag then
      if Active then
        Map.Entries[I].Activate
      else
        Map.Entries[I].Deactivate;
end;

//}}}
//{{{ TEntrySave
procedure TEntrySave.Stream;
begin
  inherited;
  if Loading then
    FTexture := Core.Graph.LoadTex(Data.ReadStr)
  else
    Data.WriteStr(Core.Graph.GetTexFileName(FTexture));
  Stream(@FSaved, SizeOf(FSaved));
end;

constructor TEntrySave.Create;
begin
  inherited Create(Map, Name, DMath.Rect(Vec2f(X, Y), Vec2f(X+128, Y+128)));
  FTexture := Core.Graph.LoadTex('DATA\textures\CycleKoh.bit');
  FSaved := False;
end;

procedure TEntrySave.DoActivate;
begin
  if not FSaved then begin
    FSaved := True;
    Core.Sound.PlaySound('DATA\sounds\SaveGame.wav');
    Game.SaveGame('DATA\saves\' + Name + '.sav');
  end;
end;

procedure TEntrySave.Render;
begin
  with Core.Graph do begin
    SetBlendMode(bmAlpha);
    if FSaved then
      SetColor(0.8, 0.4, 0.8, 1)
    else
      SetColor(0.7, 0.7, 0.7, 1);
    SetTexture(FTexture);
    DrawTexture(Rect);
    SetTexture(-1);
  end;
end;
//}}}
//{{{ TEntryRandomRespawn
procedure TEntryRandomRespawn.Stream;
begin
  inherited;
  Stream(FEPS);
end;

constructor TEntryRandomRespawn.Create(Rect: TRect; EPS: Single);
begin
  inherited Create(Game.Map, 'TEntryRandomRespawn', Rect);
  FEPS := EPS;
end;

procedure TEntryRandomRespawn.Put;
begin
end;

procedure TEntryRandomRespawn.Update;
begin
  inherited;
  if Random(10000) < FEPS then begin
    Put(Random*(Rect.A - Rect.B) + Rect.B);
  end; 
end;
//}}}
//{{{ TEntryRandomRespawn
procedure TEntryWheelRespawn.Stream; 
begin
  inherited;
  Stream(FLevel);
  Stream(@FVel, SizeOf(FVel));
end;

constructor TEntryWheelRespawn.Create;
begin
  inherited Create(Game.Map, Name, Rect);
  FVel := Vel;
  FLevel := Level;
end;

procedure TEntryWheelRespawn.DoActivate;//{{{
  var
    Wheel: TWheel;
begin
  Game.AddEnemy(TWheel.Create((Rect.A + Rect.B)/2, FLevel, FVel));
end;
//}}}
//}}}
procedure TSimpleRespawn.Put(Pos: TVec2f);//{{{
  var
    Enemy: TEnemy;
    R: Single;
begin
  Enemy := TSimpleEnemy.Create;
  Enemy.Pos := Pos;
  R := Random*2*pi;
  Enemy.Color := Vec4f(Cos(R), R/2/pi, Sin(R), 1);
  Game.AddEnemy(Enemy);
end;
//}}}
// TEntryRespawnRect//{{{
procedure TEntryRespawnRect.Stream;
begin
  inherited;
  Stream(@FWidth, SizeOf(FWidth));
  Stream(@FHeight, SizeOf(FHeight));
  Stream(@FVel, SizeOf(FVel));
end;

constructor TEntryRespawnRect.Create(Name: String; Pos, Vel: TVec2f; W, H: Integer);
begin
  inherited Create(Game.Map, Name, DMath.Rect(Pos, Pos + Vec2f(10, 10)));
  FWidth := W;
  FHeight := H;
  FVel := Vel;
end;

procedure TEntryRespawnRect.DoActivate;
  var
    Enemy: TEnemy;
    I, J: Integer;
begin
 // SDebug.Log('TEntryRespawnRect.DoActivate;');
  for I := 0 to FWidth - 1 do
    for J := 0 to FHeight - 1 do begin
      Enemy := TLineEnemy.Create;
      Enemy.Pos := Rect.A + Vec2f(I*22, J*22);
      Enemy.Vel := FVel;
      Enemy.Color := (I*Vec4f(0.8, 0.2, 0.2, 1.0) + 
                      (FWidth-I-1)*Vec4f(0.2, 0.8, 0.2, 1.0))*(1/FWidth);
      Game.AddEnemy(Enemy);
    end;
end;
//}}}
// TEntryTimer {{{
procedure TEntryTimer.Stream;
begin
  inherited;
  Stream(@FInterval, SizeOf(FInterval));
  Stream(FTag);
  Stream(FTime);
end;

constructor TEntryTimer.Create(Interval: Integer; Tag: String);
begin
  inherited Create(Game.Map, Tag, DMath.Rect(Vec2f(0, 0), Vec2f(0, 0)));
  FInterval := Interval;
  FTag := Tag;
  FTime := Game.Time;
end;

procedure TEntryTimer.Update;
  var
    I: Integer;
begin
  if FTime + FInterval < Game.Time then begin
    FTime := Game.Time;
    Activate;
  end else
    Deactivate;
  for I := 0 to Map.EntriesCount - 1 do
    if Map.Entries[I].Name = FTag then
      if Active then
        Map.Entries[I].Activate
      else
        Map.Entries[I].Deactivate;
end;
//}}}
//{{{ TEntryBossDoj
procedure TEntryBossDoj.Stream;
begin
  inherited;
  Stream(@FCreated, SizeOf(FCreated));
end;

constructor TEntryBossDoj.Create;
begin
  inherited;
  FCreated := False;
end;

procedure TEntryBossDoj.DoActivate;//{{{
  var
    Boss: TBossDoj;
begin
  if FCreated then
    Exit;
  FCreated := True;
  Boss := TBossDoj.Create((Rect.A + Rect.B)/2);
  Game.AddEnemy(Boss);
end;
//}}}
//}}}
procedure TEntryTotalWin.DoActivate;//{{{
begin
  Game.GoToScene('title');
end;
//}}}
//{{{ TMap
procedure TMap.Stream;//{{{
  var
    I, Count: Integer;
begin
  inherited;
  Count := Length(FBlocks);
  Stream(@Count, SizeOf(Count));
  SetLength(FBlocks, Count);
  for I := 0 to Count - 1 do 
    with FBlocks[I] do begin
      if Loading then
        Tex := Core.Graph.LoadTex(Data.ReadStr)
      else
        Data.WriteStr(Core.Graph.GetTexFileName(Tex));
      Stream(@Rect, SizeOf(Rect));
      Stream(@Tile, SizeOf(Tile));
      Stream(@Blend, SizeOf(Blend));
      Stream(@Color, SizeOf(Color));
      Stream(@Pos1, SizeOf(Pos1));
      Stream(@Pos2, SizeOf(Pos2));
      Stream(@Speed, SizeOf(Speed));
    end;

  Count := Length(FEntries);
  Stream(@Count, SizeOf(Count));
  SetLength(FEntries, Count);
  for I := 0 to Count - 1 do
    Stream(TStreamable(FEntries[I]));
  if Loading then
    FHeap := Core.Graph.LoadTex(Data.ReadStr)
  else begin
    Data.WriteStr(Core.Graph.GetTexFileName(FHeap));
  end;
end;
//}}}
function TMap.GetBlock(Index: Integer): TBlock;//{{{
begin
  Result := FBlocks[Index];
end;
//}}}
function TMap.GetEntry(Index: Integer): TEntry;//{{{
begin
  Result := FEntries[Index];
end;
//}}}
procedure TMap.SortBlocks;//{{{
  function Swap(A, B: Integer): Boolean;//{{{
    var
      Temp: TBlock;
  begin
    Result := False;
    if FBlocks[A].Rect.A.Y > FBlocks[B].Rect.A.Y then begin
      Temp := FBlocks[A];
      FBlocks[A] := FBlocks[B];
      FBlocks[B] := Temp;
      Result := True;
    end;
  end;//}}}
  var
    I: Integer;
    Swapped: Boolean;
begin
  Swapped := True;
  while Swapped do begin
    Swapped := False;
    for I := 0 to High(FBlocks) - 1 do
      Swapped := Swapped or Swap(I, I + 1);
    if not Swapped then
      Exit;
    Swapped := False;
    for I := High(FBlocks) - 1 downto 0 do
      Swapped := Swapped or Swap(I, I + 1);
  end;
end;
//}}}
constructor TMap.Create(Data: TData);//{{{
  var
    I: Integer;
begin
  SetLength(FSnow, 100);
  //SDebug.Log('TMap.Create');
  {SetLength(FBlocks, Data.ReadInt);
  for I := 0 to High(FBlocks) do
    with Core.Graph, FBlocks[I], Data do begin
      Tex := LoadTex(ReadStr);
      ReadVar(@Rect, SizeOf(Rect));
      ReadVar(@Tile, SizeOf(Tile));
      Blend := TBlendMode(ReadInt);
      ReadVar(@Color, SizeOf(Color));
    end;}
  //Sdebug.Log('TMap.Create done');
end;
//}}}
{procedure TMap.LoadBlocks(Blocks: array of TBlockData);//{{{
  var
    I: Integer;
begin
  SetLength(FBlocks, Length(Blocks));
  for I := 0 to High(FBlocks) do 
    with Core.Graph do begin
      FBlocks[I].Tex := LoadTex(Blocks[I].Tex);
      FBlocks[I].Rect := Blocks[I].Rect;
      FBlocks[I].Tile := Blocks[I].Tile;
      FBlocks[I].Blend := Blocks[I].Blend;
      FBlocks[I].Color := Blocks[I].Color;
      FBlocks[I].Pos1 := FBlocks[I].Rect.A;
      FBlocks[I].Pos2 := Blocks[I].Pos2;
      FBlocks[I].Speed := Blocks[I].Speed;
    end;
end;}
//}}}
procedure TMap.LoadEntries;//{{{
  var
    I: Integer;
    CurLevel: TLevelData;
begin
  with CurLevel do begin
    SetLength(FEntries, Length(Entries));
    for I := 0 to High(Entries) do 
      with Entries[I] do
        case _Type of
          etJump2Pattern: 
            FEntries[I] := TEntryJump2Pattern.Create(Self, Name, Rect, Pattern);
          etTex: FEntries[I] := TEntryTexture.Create(Self, Name, Rect, Blend, Vec4f(1, 1, 1, 1));
          etTag: FEntries[I] := TEntryTag.Create(Self, Name, Rect, Tag);
        end;
  end;
  FHeap := -1;
  { TODO if Game.GetLevel.Heap <> '' then
    FHeap := Core.Graph.LoadTex(Game.GetLevel.Heap);}
end;
//}}}
procedure TMap.Render;//{{{
  var
    I: Integer;
begin
  Core.Graph.SetLineWidth(1);
  for I := 0 to High(FBlocks) do 
    with Core.Graph, FBlocks[I] do begin
      SetBlendMode(Blend);
      //Color4f(Color);
      SetColor(1, 1, 1, 1);
      SetTexture(Tex);
      DrawTiles(Rect, Tile);
      if FHeap <> -1 then begin
        SetTexture(FHeap);
        SetBlendMode(bmAlpha);
        Color4f(FHeapColor);
        DrawTiles(DMath.Rect(Rect.A, Vec2f(Rect.B.X, Rect.A.Y+20)), 
                  Vec2f(128, 128));
      end;
      SetBlendMode(bmNo);
      SetTexture(-1);
      SetColor(0, 0, 0, 1);
      DrawRect(Rect, False);
    end;
  Core.Graph.Color4f(FHeapColor);
  for I := 0 to High(FSnow) do begin
    Core.Graph.DrawCircle(FSnow[I].Pos + Vec2f(0, GetTimer/100), 5);
  end;
  Core.Graph.SetBlendMode(bmAlpha);
  Core.Graph.SetColor(1, 1, 1, 1);
  Core.Graph.SetTexture(-1);
  for I := 0 to High(FEntries) do
    FEntries[I].Render;
  Core.Graph.SetBlendMode(bmNo);
  Core.Graph.SetColor(1, 1, 1, 1);
  Core.Graph.SetTexture(-1);
end;
//}}}
procedure TMap.Update;//{{{
  var
    I: Integer;
    Size: TVec2f;
begin
  for I := 0 to High(FBlocks) do
    if FBlocks[I].Speed <> 0.0 then 
      with FBlocks[I] do begin
        Size := Rect.B - Rect.A;
        Rect.A := Pos1 + (Sin(Game.Time/1000*Speed)+1)/2*(Pos2 - Pos1);
        Rect.B := Rect.A + Size;
      end;
  if Random(10) = 0 then begin
    SetLength(FSnow, Length(FSnow) + 1);
    FSnow[High(FSnow)].Pos := Vec2f(Game.Player.Pos.X ,//+ Random(1600)-800,
                                    Game.Player.Pos.Y - 300);
    FSnow[High(FSnow)].Start := GetTimer;
  end;
  for I := 0 to High(FSnow) do
    if GetTimer - FSnow[I].Start > 30000 then begin
      FSnow[I].Pos := Vec2f(Game.Player.Pos.X + Random(1600)-800,
                            Game.Player.Pos.Y - 600);
      FSnow[I].Start := GetTimer;
    end;

        //SortBlocks;
end;
//}}}
procedure TMap.UpdateEntries(Player: TRect);//{{{
  var
    I: Integer;
begin
  for I := 0 to High(FEntries) do
    FEntries[I].Update;
end;
//}}}
function TMap.Collide(Rect: TRect; //{{{
                      Block: Integer; 
                      var Delta: TVec2f): TMapCollision;
  var
    Other: TRect;
begin
  Other := FBlocks[Block].Rect;
  Delta := Vec2f(0, 0);
  if not Intersect(Rect, Other) then begin
    Result := mcNo
  end else if (Rect.B.Y >= Other.A.Y)and(Rect.A.Y < Other.A.Y) then begin
    Result := mcUp;
    Delta := Vec2f(0, Other.A.Y - Rect.B.Y);
  end else if Rect.A.X < Other.A.X then begin
    Result := mcSide;
    Delta := Vec2f(Other.A.X - Rect.B.X, 0);
  end else if Rect.B.X > Other.B.X then begin
    Result := mcSide;
    Delta := Vec2f(Other.B.X - Rect.A.X, 0);
  end else begin
    Result := mcUp;
    Delta := Vec2f(0, Other.A.Y - Rect.B.Y);
  end;
end;
//}}}
function TMap.Higher(Rect: TRect; Block: Integer): Boolean;//{{{
begin
  Result := Rect.B.Y <= FBlocks[Block].Rect.A.Y + 10;
end;
//}}}
function TMap.GetLevel(Rect: TRect): Integer;//{{{
begin
  Result := 0;
  if Rect.B.Y <= FBlocks[0].Rect.A.Y then
    Exit;
  while (Result < Length(FBlocks)) and (FBlocks[Result].Rect.A.Y < Rect.B.Y) do     
    Inc(Result);
end;
//}}}
procedure TMap.AddBlock(Block: TBlock);//{{{
begin
  SetLength(FBlocks, Length(FBlocks) + 1);
  FBlocks[High(FBlocks)] := Block;
end;

//}}}
procedure TMap.AddEntry;//{{{
begin
  SetLength(FEntries, Length(FEntries) + 1);
  FEntries[High(FEntries)] := Entry;
end;
//}}}
function TMap.BlocksCount: Integer; //{{{
begin
  Result := Length(FBlocks);
end;
//}}}
function TMap.EntriesCount: Integer; //{{{
begin
  Result := Length(FEntries);
end;
//}}}
procedure TMap.XMLBlocks(Exe: TXMLExecutor; Item: PXMLItem; Into: Boolean);
  var
    I: Integer;
begin
  if Into then begin
    Exe.AddTag('block', XMLBlock);
    FXMLPos := Vec2f(0, 600);
    FXMLSize := Vec2f(800, 200);
    FXMLTex := 'data\tiles\RombGreen.tga';
    FXMLHeap := 'data\tiles\Heap.bit';
    FXMLBlend := bmAlpha;
    FXMLColor := Vec4f(0.5, 1, 0, 1); 
    FXMLTile := Vec2f(64, 64);
    FHeapColor := Vec4f(72/255, 178/255, 129/255, 1.0);
    for I := 0 to High(Item.Params) do
      with Item.Params[I] do
        if Name = 'heap' then
          FXMLHeap := Value
        else if Name = 'heapcolor' then
          FHeapColor := StrToColor(Value);
    SetLength(FBlocks, 0);
  end else begin
    FHeap := Core.Graph.LoadTex(FXMLHeap);
  end;
end;

procedure TMap.XMLBlock(Exe: TXMLExecutor; Item: PXMLItem; Into: Boolean);
  var
    I: Integer;
begin
  for I := 0 to High(Item.Params) do
    with Item.Params[I] do
      if Name = 'x' then
        FXMLPos.X := StrToFloat(Value)
      else if Name = 'y' then
        FXMLPos.Y := StrToFloat(Value)
      else if Name = 'w' then
        FXMLSize.X := StrToFloat(Value)
      else if Name = 'h' then
        FXMLSize.Y := StrToFloat(Value)
      else if Name = 'tex' then
        FXMLTex := Value
      else if Name = 'blend' then begin
        //
      end else if Name = 'color' then
        FXMLColor := StrToColor(Value)
      else if Name = 'tilex' then
        FXMLTile.X := StrToFloat(Value)
      else if Name = 'tiley' then
        FXMLTile.Y := StrToFloat(Value);
  SetLength(FBlocks, Length(FBlocks) + 1);
  with Core.Graph do begin
    FBlocks[High(FBlocks)].Tex := LoadTex(FXMLTex);
    FBlocks[High(FBlocks)].Rect.A := FXMLPos;
    FBlocks[High(FBlocks)].Rect.B := FXMLPos + FXMLSize;
    FBlocks[High(FBlocks)].Tile := FXMLTile;
    FBlocks[High(FBlocks)].Blend := FXMLBlend;
    FBlocks[High(FBlocks)].Color := FXMLColor;
    FBlocks[High(FBlocks)].Pos1 := FXMLPos;
    FBlocks[High(FBlocks)].Pos2 := FXMLPos;
    FBlocks[High(FBlocks)].Speed := 0;
  end;
end;

//}}}

initialization
  Factory.Regist(TEntry);
  Factory.Regist(TEntryTag);
  Factory.Regist(TEntryTexture);
  Factory.Regist(TEntryText);
  Factory.Regist(TEntryJump2Pattern);
  Factory.Regist(TEntryMusic);
  Factory.Regist(TEntryTag);
  Factory.Regist(TEntrySave);
  Factory.Regist(TEntryBossDoj);
  Factory.Regist(TEntryRandomRespawn);
  Factory.Regist(TSimpleRespawn);
  Factory.Regist(TEntryRespawnRect);
  Factory.Regist(TEntryWheelRespawn);
  Factory.Regist(TEntryTimer);
  Factory.Regist(TEntryTotalWin);
  Factory.Regist(TMap);
end.
