unit e_game;

interface

uses OpenGL, windows, e_models, e_textures, e_types, e_global, e_utils,
     e_particles, e_sound, e_camera, e_font, e_weapons,
     e_player, e_about, inifiles, e_lines,
     e_console, e_enemies, e_towers, e_menu,
     e_level1, e_level2, e_level3, e_level4, e_level5, e_level6;

type
  TBonus = class(TCreature)
    mode   : byte;
    dead   : boolean;
    timer  : single;
  public
    constructor Create(_mode : byte; _pos : TPoint; _angle : single);
    procedure   RenderBack;
    procedure   RenderFront;
    procedure   Update(count : single);
  end;

  TLabel = class
    text  : string;
    pos   : TPoint;
    size  : single;
    alpha : single;
    alpha_target : single;
    gui   : boolean;
  public
    constructor Create(_text : string; _x : single = 0; _y : single = 0; _size : single = 20;  _gui : boolean = true);
    procedure   Render;
    procedure   Update(count : single);
    procedure   Show;
    procedure   Hide;
  end;

  TStageObject = record
    mode   : byte;
    pos    : TPoint;
  end;

  TGame = class
    player         : TPlayer;
    towers         : array of TTower;
    bonuses        : array of TBonus;
    bonuses_active : boolean;
    bullets        : array of TChuck;
    enemies        : array of TEnemy;
    labels         : array of TLabel;
    objects_number : integer;
    objects        : array of TObject;

    stage          : array of TStageObject;
    stage_number   : integer;
    stage_number_next : integer;
    stage_progress : integer;

    timer_wait     : single;
    timer_bonus    : single;

    script_waiting : integer;
    script         : integer;

    game_over      : boolean;

    play_time      : single;

    label_game_over1 : TLabel;
    label_game_over2 : TLabel;
    label_free     : TLabel;
    label_health   : TLabel;
    label_angles   : TLabel;
    label_fps      : TLabel;
  public
    constructor Create(_stage_number : integer);

    procedure   DebugUpdate;

    procedure   StageClear;
    procedure   StageUpdate(count : single);
    procedure   StageSkip(full : boolean = false);

    function    AddObject(x, y, angle, sx, sy : single; texture : gluint) : TObject;
    procedure   CountObjects;
    function    AddTower(mode : byte; pos : TPoint; angle : single) : TTower;
    function    AddBonus(mode : byte; pos : TPoint; angle : single) : TBonus;
    function    AddBullet(mode : byte; parent : TCreature) : TChuck;
    function    AddEnemy(mode : byte; pos : TPoint) : TEnemy;
    function    AddLabel(_text : string; _x : single = 0; _y : single = 0; _size : single = 20; _gui : boolean = false) : TLabel;

    function    CollideMap(creature : TCreature) : boolean;
    procedure   Explosion(pos : TPoint; radius, power : single; sound : boolean = true; distance : single = -1);
    
    procedure   Render;
    function    Wait(seconds : single) : boolean;
    procedure   UpdateLabels;
    procedure   Update(count : single);
    procedure   Save;
    procedure   GameOver(win : boolean);
  end;

  TProfiles = class
    procedure Load;
    procedure Save;
  end;

var
  game : TGame;
  profiles : TProfiles;

implementation

uses e_main;

constructor TBonus.Create(_mode : byte; _pos : TPoint; _angle : single);
begin
mode      := _mode;
pos       := _pos;
angle     := _angle;
size      := 0.2;
dead      := false;
timer     := 300;
end;

procedure TBonus.RenderBack;
var
t : single;
begin
glBlendFunc(GL_SRC_ALPHA, GL_ONE);
SetTexture(texture_bonus);
glColor4f(1, 1, 1, 0.3);

glPushMatrix;
with pos do glTranslatef(x, y, 0);

t := timer / 200; if t > 1 then t := 1;
glScalef(size * 2.2 * t, size * 2.2 * t, 1);


glCallList(model_quad);
glPopMatrix;
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
end;

procedure TBonus.RenderFront;
var
t : single;
begin
SetTexture(texture_bonus);
glColor4f(1, 0.8, 0, 1);

glPushMatrix;
with pos do glTranslatef(x, y, 0);

t := timer / 200; if t > 1 then t := 1;
glScalef(size * 2 * t, size * 2 * t, 1);

glCallList(model_quad);
glPopMatrix;
end;

procedure TBonus.Update(count : single);
begin
timer := timer - count;
if timer <= 0 then
  begin
    dead := true;
    exit;
  end;

if not CollideCreatures(self, game.player) then
  exit;

SoundPlay(sound_bonus, pos.x);

game.player.GetScore(1);

dead := true;
particles.AddEmitter(particles_bonus, pos, 10, GetColor(1, 0.8, 0, 1)).layer := 1;
end;

constructor TLabel.Create(_text : string; _x : single = 0; _y : single = 0; _size : single = 20;  _gui : boolean = true);
begin
text          := _text;
pos           := GetPoint(_x, _y);
size          := _size;
gui           := _gui;
alpha         := -0.1;
alpha_target  := -0.1;
end;

procedure TLabel.Render;
begin
if alpha < 0 then
  exit;
  
glColor4f(1, 1, 1, sqr(alpha));
if gui then
  font.TextOut(size * alpha / 1.5, camera.eye.x + pos.x, camera.eye.y + pos.y, text, faCenter)
else
  font.TextOut(size * alpha / 1.5, pos.x, pos.y, text, faCenter);
end;

procedure TLabel.Update(count : single);
begin
SqrAspire(alpha, alpha_target, count * 0.004);
end;

procedure TLabel.Show;
begin
alpha_target := 1.5;
end;

procedure TLabel.Hide;
begin
alpha_target := -0.1;
end;

constructor TGame.Create(_stage_number : integer);
begin
stage_number := _stage_number;
menu_selected := stage_number;
menu_selected_pos := menu_selected;
player := TPlayer.Create;

camera.eye := GetPoint(0, 0);
camera.target := camera.eye;

clear_color := GetColor(0, 0, 0, 1);

bonuses_active  := false;

timer_wait      := 0;
timer_bonus     := 0;
script_waiting  := -1;
script          := 0;

objects_number  := 0;

game_over       := false;

label_game_over1 := AddLabel('', 0, 1, 5.15, true);
label_game_over2 := AddLabel('', 0, 0, 10, true);

label_free    := AddLabel('free', 11, -7, 5, true);
label_fps    := AddLabel('fps', 11, -10, 5, true);
label_angles := AddLabel('a: ', 11, -9, 5, true);
label_health := AddLabel('health', -11, -8, 10, true);
end;

procedure TGame.DebugUpdate;
begin
label_health.Hide;
label_free.Hide;
label_fps.Hide;
label_angles.Hide;

if not render_debug then exit;

label_health.Show;
label_free.Show;
label_fps.Show;
label_angles.Show;

label_health.text   := 'health ' + inttostr(player.health);
label_fps.text      := 'fps ' + inttostr(fps);
label_angles.text   := 'a: ' + inttostr(round(player.angle));
end;

procedure TGame.StageClear();
var
i : integer;
begin
stage_progress := 0;
script := 0;

for i := 0 to high(enemies) do
  if enemies[i] <> nil then
    begin
      enemies[i].Free;
      enemies[i] := nil;
    end;
SetLength(enemies, 0);

for i := 0 to high(enemies) do
  if enemies[i] <> nil then
    begin
      enemies[i].Free;
      enemies[i] := nil;
    end;
SetLength(enemies, 0);

for i := 0 to high(bullets) do
  if bullets[i] <> nil then
    begin
      bullets[i].Free;
      bullets[i] := nil;
    end;
SetLength(bullets, 0);

for i := 0 to high(bonuses) do
  if bonuses[i] <> nil then
    begin
      bonuses[i].Free;
      bonuses[i] := nil;
    end;
SetLength(bonuses, 0);
end;

procedure TGame.StageUpdate(count : single);
begin
case stage_number of
  1 : e_level1.StageUpdate(count);
  2 : e_level2.StageUpdate(count);
  3 : e_level3.StageUpdate(count);
  4 : e_level4.StageUpdate(count);
  5 : e_level5.StageUpdate(count);
  6 : e_level6.StageUpdate(count);
end;

if stage_progress > high(stage) then
  exit;

{with stage[stage_progress] do
  if pos < camera.eye.x + game_width then
    begin
      if mode < 10 then
        game.AddBonus(mode, pos);
      if stage[stage_progress].mode >= 10 then
        game.AddEnemy(mode - 10, pos);

      inc(stage_progress);
    end;   }
end;

procedure TGame.StageSkip(full : boolean = false);
begin
case stage_number of
  1 : e_level1.ScriptSkip(full);
  2 : e_level2.ScriptSkip(full);
  3 : e_level3.ScriptSkip(full);
  4 : e_level4.ScriptSkip(full);
  5 : e_level5.ScriptSkip(full);
  6 : e_level6.ScriptSkip(full);
end;
end;

function TGame.AddObject(x, y, angle, sx, sy : single; texture : gluint) : TObject;
begin
objects[objects_number] := TObject.Create(GetPoint(x, y), angle, GetPoint(sx, sy), texture);
result := objects[objects_number];
inc(objects_number);
end;

procedure TGame.CountObjects;
var
i : integer;
begin
for i := 0 to high(objects) do
  if objects[i] <> nil then
    objects[i].Count;
end;

function TGame.AddTower(mode : byte; pos : TPoint; angle : single) : TTower;
var
i, max : integer;
begin
result := nil;

if game_over then exit;

i   := 0;
max := high(towers);
while (i <= max) and (towers[i] <> nil) do inc(i);
if i > max then SetLength(towers, i + 1);

if towers[i] <> nil then towers[i].Free;

towers[i] := TTower.Create(mode, pos, angle);

result := towers[i];
end;

function TGame.AddBonus(mode : byte; pos : TPoint; angle : single) : TBonus;
var
i, max : integer;
begin
result := nil;

if game_over then exit;

i   := 0;
max := high(bonuses);
while (i <= max) and (bonuses[i] <> nil) do inc(i);
if i > max then SetLength(bonuses, i + 1);

if bonuses[i] <> nil then bonuses[i].Free;

bonuses[i] := TBonus.Create(mode, pos, angle);

result := bonuses[i];
end;

function TGame.AddBullet(mode : byte; parent : TCreature) : TChuck;
var
i, max : integer;
move : single;
begin
result := nil;

i   := 0;
max := high(bullets);
while (i <= max) and (bullets[i] <> nil) do inc(i);
if i > max then SetLength(bullets, i + 1);

if bullets[i] <> nil then bullets[i].Free;
bullets[i] := WeaponShoot(mode, parent);

result := bullets[i];
end;

function TGame.AddEnemy(mode : byte; pos : TPoint) : TEnemy;
var
i, max : integer;     
begin
result := nil;

if game_over then exit;

i   := 0;
max := high(enemies);
while (i <= max) and (enemies[i] <> nil) do inc(i);
if i > max then SetLength(enemies, i + 1);

if enemies[i] <> nil then enemies[i].Free;

case mode of
  1 : enemies[i] := TEnemy1.Create(pos);
  2 : enemies[i] := TEnemy2.Create(pos);
  3 : enemies[i] := TEnemy3.Create(pos);
  else enemies[i] := TEnemy1.Create(pos);
end;

result := enemies[i];
end;

function TGame.AddLabel(_text : string; _x : single = 0; _y : single = 0; _size : single = 20; _gui : boolean = false) : TLabel;
var
i, max : integer;     
begin
result := nil;

if game_over then exit;

i   := 0;
max := high(labels);
while (i <= max) and (labels[i] <> nil) do inc(i);
if i > max then SetLength(labels, i + 1);

if labels[i] <> nil then labels[i].Free;

labels[i] := TLabel.Create(_text, _x, _y, _size, _gui);

result := labels[i];
end;


function TGame.CollideMap(creature : TCreature) : boolean;
var
i, j : integer;
s : single;
begin
result := true;

s := creature.size * 0.5;
for i := 0 to high(objects) do
  if objects[i] <> nil then
    if objects[i].collide_units then
      for j := 0 to high(objects[i].lines) do
        if Distance(creature.pos, objects[i].lines[j].a, objects[i].lines[j].b) < s then
          exit;

result := false;
end;

procedure TGame.Explosion(pos : TPoint; radius, power : single; sound : boolean = true; distance : single = -1);
var
i : integer;
begin
particles.AddEmitter(particles_explosion, pos, 15, GetColor(1, 1, 1, 1));
if sound then
  SoundPlay(sound_explosion, pos.x);

if distance < 0 then
  distance := PointsDistance(pos, game.player.pos);

if distance <= radius then
  game.player.GetDamage(power, distance);

for i := 0 to high(game.bullets) do
  if game.bullets[i] <> nil then
    if not game.bullets[i].dead then
      if PointsDistance(pos, game.bullets[i].pos) < radius then
        begin
          game.bullets[i].dead := true;
          with game.bullets[i] do
            game.Explosion(pos, radius, power, false);
        end;
//game.player.angle := game.player.angle - PointsAngle(pos, game.player.pos) * 0.3;
//game.player.speed.x := game.player.speed.x + 0.1 * (game.player.pos.x - pos.x) / distance;
//game.player.speed.y := game.player.speed.y + 0.1 * (game.player.pos.y - pos.y) / distance;
end;

procedure TGame.Render;
var
i, j : integer;
s : single;
p, n : TPoint;
begin
glDisable(GL_DEPTH_TEST);

case stage_number of
  1 : e_level1.RenderBack;
  2 : e_level2.RenderBack;
  3 : e_level3.RenderBack;
  4 : e_level4.RenderBack;
  5 : e_level5.RenderBack;
  6 : e_level6.RenderBack;
end;
for i := 0 to high(labels) do
  if labels[i] <> nil then
    if not labels[i].gui then
      labels[i].Render;
case stage_number of
  1 : e_level1.RenderFront;
  2 : e_level2.RenderFront;
  3 : e_level3.RenderFront;
  4 : e_level4.RenderFront;
  5 : e_level5.RenderFront;
  6 : ;//  
end;

particles.Render(1);

for i := 0 to high(bonuses) do
  if bonuses[i] <> nil then
    bonuses[i].RenderFront;
for i := 0 to high(bonuses) do
  if bonuses[i] <> nil then
    bonuses[i].RenderBack;

for i := 0 to high(towers) do
  if towers[i] <> nil then
    towers[i].RenderBack;

particles.Render(2);

for i := 0 to high(enemies) do
  if enemies[i] <> nil then
    enemies[i].RenderBack;

player.Render;

for i := 0 to high(bullets) do
  if bullets[i] <> nil then
    bullets[i].Render;

for i := 0 to high(enemies) do
  if enemies[i] <> nil then
    enemies[i].RenderFront;

for i := 0 to high(towers) do
  if towers[i] <> nil then
    towers[i].RenderFront;

for i := 0 to high(objects) do
  if objects[i] <> nil then
    objects[i].Render;

particles.Render(3);
particles.Render(4);

case stage_number of
  //  
  6 : e_level6.RenderFront;
end;

for i := 0 to high(labels) do
  if labels[i] <> nil then
    if labels[i].gui then
      labels[i].Render;

if render_debug then
  begin
    // 
    label_free.text := inttostr(round(mouse_game.x)) + 'x' + inttostr(round(mouse_game.y));
    SetTexture(texture_bonus);
    glColor4f(1, 0, 0, 1);
    glPushMatrix;
    with mouse_game do glTranslatef(x, y, 0);
    glScalef(0.3, 0.3, 1);
    glCallList(model_quad);
    glPopMatrix;
  end;
end;

function TGame.Wait(seconds : single) : boolean;
begin
if script = script_waiting then
  begin
    result := timer_wait < 0;
    exit;
  end;
script_waiting := script;
timer_wait := seconds * 1000;
result := false;
end;

procedure TGame.UpdateLabels;
begin
{
with GUIAddLabel(20, 55, 200, 200, 'score ')^ do
  target := @game.player.score;

if not render_debug then
  exit;

with GUIAddLabel(820, 630, 120, 120, 'score     ')^ do
  target := @game.player.score;
with GUIAddLabel(820, 650, 120, 120, 'life      ')^ do
  target := @game.player.life;
with GUIAddLabel(820, 670, 120, 120, 'pos       ')^ do
  target := @game.player.pos_i;
with GUIAddLabel(820, 690, 120, 120, 'map       ')^ do
  target := @game.player.map;
with GUIAddLabel(820, 710, 120, 120, 'stage     ')^ do
  target := @game.stage_progress;
with GUIAddLabel(820, 730, 120, 120, 'script    ')^ do
  target := @game.script;
with GUIAddLabel(820, 750, 120, 120, 'fps       ')^ do
  target := @fps;      }
end;

procedure TGame.Update(count : single);
var
i, j : integer;
d, s : single;
begin
game_played := true;

count := count * 0.5; //  /. ,   .
DebugUpdate;

if timer_wait >= 0 then
  timer_wait := timer_wait - count;

if bonuses_active then
  begin    
    timer_bonus := timer_bonus - count;
    if timer_bonus < 0 then
      begin
        timer_bonus := 200;
        with player do
          AddBonus(1, GetPoint(pos.x + cos(angle / 180 * pi) * 3, pos.y + sin(angle / 180 * pi) * 3), player.angle);
      end;
  end;


UpdateLabels;
for i := 0 to high(labels) do
  if labels[i] <> nil then
    labels[i].Update(count);

StageUpdate(count);

for i := 0 to high(objects) do
  if objects[i] <> nil then
    objects[i].Update(count);

player.Update(count);

for i := 0 to high(bullets) do
  if bullets[i] <> nil then
    with bullets[i] do
      if not dead then
        Update(count)
      else
        begin
          Free;
          bullets[i] := nil;
        end;
        
for i := 0 to high(towers) do
  if towers[i] <> nil then
    with towers[i] do
      Update(count);

for i := 0 to high(enemies) do
  if enemies[i] <> nil then
    with enemies[i] do
      if not dead then
        Update(count)
      else
        begin
          if particles1 <> nil then particles1.loop := false;
          if particles2 <> nil then particles2.loop := false;
          Free;
          enemies[i] := nil;
        end; 

for i := 0 to high(bonuses) do
  if bonuses[i] <> nil then
    with bonuses[i] do
      if not dead then
        Update(count)
      else
        begin
          Free;
          bonuses[i] := nil;
        end;

particles.Update(count);
end;

procedure TGame.Save;
begin

end;

procedure TGame.GameOver(win : boolean);
begin
if game_over then exit;

if win then
  begin
    label_game_over1.text := 'press spacebar to';
    label_game_over2.text := 'next level';
    stage_number_next := stage_number + 1;
  end
else
  begin
    label_game_over1.text := 'press spacebar to';
    label_game_over2.text := 'try again';
    stage_number_next := stage_number;
  end;

if stage_number_next > max_levels then
  begin
    label_game_over1.text := 'thanks for playing';
    label_game_over2.text := 'good luck';
    stage_number_next := 0;
  end;

label_game_over1.Show;
label_game_over2.Show;

game_over := true;
game_played := false;

Save;
end;

procedure TProfiles.Load;
begin
end;

procedure TProfiles.Save;
begin
end;

end.
